# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import random
from collections.abc import Mapping
from typing import Any
from unittest import mock

import numpy as np
import pandas as pd
from ax.analysis.analysis_card import AnalysisCard
from ax.analysis.plotly.parallel_coordinates import ParallelCoordinatesPlot
from ax.api.client import Client
from ax.api.configs import ChoiceParameterConfig, RangeParameterConfig, StorageConfig
from ax.api.protocols.metric import IMetric
from ax.api.protocols.runner import IRunner
from ax.api.types import TParameterization
from ax.core.experiment import Experiment
from ax.core.formatting_utils import DataType
from ax.core.map_data import MapData
from ax.core.map_metric import MapMetric
from ax.core.objective import MultiObjective, Objective, ScalarizedObjective
from ax.core.optimization_config import OptimizationConfig
from ax.core.outcome_constraint import ComparisonOp, OutcomeConstraint
from ax.core.parameter import (
    ChoiceParameter,
    ParameterType as CoreParameterType,
    RangeParameter,
)
from ax.core.parameter_constraint import ParameterConstraint
from ax.core.search_space import SearchSpace
from ax.core.trial import Trial
from ax.core.trial_status import TrialStatus
from ax.early_stopping.strategies import PercentileEarlyStoppingStrategy
from ax.exceptions.core import UnsupportedError
from ax.service.utils.with_db_settings_base import (
    _save_generation_strategy_to_db_if_possible,
)
from ax.storage.sqa_store.db import init_test_engine_and_session_factory
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_experiment,
    get_branin_optimization_config,
    get_percentile_early_stopping_strategy,
)
from ax.utils.testing.mock import mock_botorch_optimize
from ax.utils.testing.modeling_stubs import get_generation_strategy
from pyre_extensions import assert_is_instance, none_throws, override


class TestClient(TestCase):
    def test_configure_experiment(self) -> None:
        client = Client()

        float_parameter = RangeParameterConfig(
            name="float_param",
            parameter_type="float",
            bounds=(0, 1),
        )
        int_parameter = RangeParameterConfig(
            name="int_param",
            parameter_type="int",
            bounds=(0, 1),
        )
        choice_parameter = ChoiceParameterConfig(
            name="choice_param",
            parameter_type="str",
            values=["a", "b", "c"],
        )

        client.configure_experiment(
            name="test_experiment",
            parameters=[float_parameter, int_parameter, choice_parameter],
            parameter_constraints=["int_param <= float_param"],
            description="test description",
            owner="miles",
        )
        self.assertEqual(
            client._experiment,
            Experiment(
                search_space=SearchSpace(
                    parameters=[
                        RangeParameter(
                            name="float_param",
                            parameter_type=CoreParameterType.FLOAT,
                            lower=0,
                            upper=1,
                        ),
                        RangeParameter(
                            name="int_param",
                            parameter_type=CoreParameterType.INT,
                            lower=0,
                            upper=1,
                        ),
                        ChoiceParameter(
                            name="choice_param",
                            parameter_type=CoreParameterType.STRING,
                            values=["a", "b", "c"],
                            is_ordered=False,
                            sort_values=False,
                        ),
                    ],
                    parameter_constraints=[
                        ParameterConstraint(
                            constraint_dict={"int_param": 1, "float_param": -1}, bound=0
                        )
                    ],
                ),
                name="test_experiment",
                description="test description",
                properties={"owners": ["miles"]},
                default_data_type=DataType.MAP_DATA,
            ),
        )

        with self.assertRaisesRegex(UnsupportedError, "Experiment already configured"):
            client.configure_experiment(
                name="test_experiment",
                parameters=[float_parameter, int_parameter, choice_parameter],
                parameter_constraints=["int_param <= float_param"],
                description="test description",
                owner="miles",
            )

    def test_configure_optimization(self) -> None:
        client = Client()

        float_parameter = RangeParameterConfig(
            name="float_param",
            parameter_type="float",
            bounds=(0, 1),
        )
        int_parameter = RangeParameterConfig(
            name="int_param",
            parameter_type="int",
            bounds=(0, 1),
        )
        choice_parameter = ChoiceParameterConfig(
            name="choice_param",
            parameter_type="str",
            values=["a", "b", "c"],
        )

        client.configure_experiment(
            name="test_experiment",
            parameters=[float_parameter, int_parameter, choice_parameter],
            parameter_constraints=["int_param <= float_param"],
            description="test description",
            owner="miles",
        )

        client.configure_optimization(
            objective="-ne",
            outcome_constraints=["qps >= 0"],
        )

        self.assertEqual(
            client._experiment.optimization_config,
            OptimizationConfig(
                objective=Objective(
                    metric=MapMetric(name="ne", lower_is_better=True), minimize=True
                ),
                outcome_constraints=[
                    OutcomeConstraint(
                        metric=MapMetric(name="qps"),
                        op=ComparisonOp.GEQ,
                        bound=0.0,
                        relative=False,
                    )
                ],
            ),
        )

        empty_client = Client()

        with self.assertRaisesRegex(AssertionError, "Experiment not set"):
            empty_client.configure_optimization(
                objective="ne",
                outcome_constraints=["qps >= 0"],
            )

        # Test that metrics not in the objective string are downgraded to tracking
        # metrics rather than being removed
        client = Client()
        client.configure_experiment(
            name="test_experiment",
            parameters=[float_parameter],
        )

        custom_metric1 = DummyMetric(name="metric1")
        custom_metric2 = DummyMetric(name="metric2")
        custom_metric3 = DummyMetric(name="metric3")
        client.configure_metrics(
            metrics=[custom_metric1, custom_metric2, custom_metric3]
        )

        # Verify all metrics are added as tracking metrics
        self.assertEqual(len(client._experiment.tracking_metrics), 3)
        self.assertIn(custom_metric1, client._experiment.tracking_metrics)
        self.assertIn(custom_metric2, client._experiment.tracking_metrics)
        self.assertIn(custom_metric3, client._experiment.tracking_metrics)

        # Configure optimization with only metric1 in the objective
        client.configure_optimization(
            objective="metric1",
        )

        # Verify metric2 and metric3 are still present as a tracking metric
        self.assertIn(custom_metric2, client._experiment.tracking_metrics)
        self.assertIn(custom_metric3, client._experiment.tracking_metrics)

        # Verify metric1 is now part of the objective and no longer a tracking metric
        optimization_config = client._experiment.optimization_config
        self.assertIsNotNone(optimization_config)
        objective = assert_is_instance(
            optimization_config.objective,
            Objective,
        )
        self.assertEqual(objective.metric.name, "metric1")

        # Verify no metrics were removed, just moved from tracking to objective
        all_metrics = [objective.metric] + list(client._experiment.tracking_metrics)
        self.assertEqual(len(all_metrics), 3)

    def test_configure_runner(self) -> None:
        client = Client()
        runner = DummyRunner()

        with self.assertRaisesRegex(AssertionError, "Experiment not set"):
            client.configure_runner(runner=runner)

        client.set_experiment(experiment=get_branin_experiment())
        client.configure_runner(runner=runner)

        self.assertEqual(client._experiment.runner, runner)

    def test_configure_metric(self) -> None:
        client = Client()
        custom_metric = DummyMetric(name="custom")

        with self.assertRaisesRegex(AssertionError, "Experiment not set"):
            client.configure_metrics(metrics=[custom_metric])

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(0, 1))
            ],
            name="foo",
        )

        # Test replacing a single objective
        client.configure_optimization(objective="custom")
        client.configure_metrics(metrics=[custom_metric])

        self.assertEqual(
            custom_metric,
            none_throws(client._experiment.optimization_config).objective.metric,
        )

        # Test replacing a multi-objective
        client.configure_optimization(objective="custom, foo")
        client.configure_metrics(metrics=[custom_metric])

        self.assertIn(
            custom_metric,
            assert_is_instance(
                none_throws(client._experiment.optimization_config).objective,
                MultiObjective,
            ).metrics,
        )
        # Test replacing a scalarized objective
        client.configure_optimization(objective="custom + foo")
        client.configure_metrics(metrics=[custom_metric])

        self.assertIn(
            custom_metric,
            assert_is_instance(
                none_throws(client._experiment.optimization_config).objective,
                ScalarizedObjective,
            ).metrics,
        )

        # Test replacing an outcome constraint
        client.configure_optimization(
            objective="foo", outcome_constraints=["custom >= 0"]
        )
        client.configure_metrics(metrics=[custom_metric])

        self.assertEqual(
            custom_metric,
            none_throws(client._experiment.optimization_config)
            .outcome_constraints[0]
            .metric,
        )

        # Test adding a tracking metric
        client = Client()  # Start a fresh Client
        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(0, 1))
            ],
            name="foo",
        )
        client.configure_metrics(metrics=[custom_metric])

        self.assertEqual(
            custom_metric,
            client._experiment.tracking_metrics[0],
        )

    def test_set_experiment(self) -> None:
        client = Client()
        experiment = get_branin_experiment()

        client.set_experiment(experiment=experiment)

        self.assertEqual(client._experiment, experiment)

    def test_set_optimization_config(self) -> None:
        client = Client()
        optimization_config = get_branin_optimization_config()

        with self.assertRaisesRegex(AssertionError, "Experiment not set"):
            client.set_optimization_config(optimization_config=optimization_config)

        client.set_experiment(experiment=get_branin_experiment())
        client.set_optimization_config(
            optimization_config=optimization_config,
        )

        self.assertEqual(client._experiment.optimization_config, optimization_config)

    def test_set_generation_strategy(self) -> None:
        client = Client()
        client.set_experiment(experiment=get_branin_experiment())

        generation_strategy = get_generation_strategy()

        client.set_generation_strategy(generation_strategy=generation_strategy)
        self.assertEqual(client._generation_strategy, generation_strategy)

    def test_set_early_stopping_strategy(self) -> None:
        client = Client()
        early_stopping_strategy = get_percentile_early_stopping_strategy()

        client.set_early_stopping_strategy(
            early_stopping_strategy=early_stopping_strategy
        )
        self.assertEqual(client._early_stopping_strategy, early_stopping_strategy)

    def test_get_next_trials(self) -> None:
        client = Client()

        with self.assertRaisesRegex(AssertionError, "Experiment not set"):
            client.get_next_trials(max_trials=1)

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
                RangeParameterConfig(name="x2", parameter_type="float", bounds=(-1, 1)),
            ],
            name="foo",
        )

        with self.assertRaisesRegex(UnsupportedError, "OptimizationConfig not set"):
            client.get_next_trials(max_trials=1)

        client.configure_optimization(objective="foo")
        client.configure_generation_strategy(
            # Set this to a large number so test runs fast
            initialization_budget=999,
        )

        # Test can generate one trial
        trials = client.get_next_trials(max_trials=1)
        self.assertEqual(len(trials), 1)
        self.assertEqual({*trials[0].keys()}, {"x1", "x2"})
        for parameter in ["x1", "x2"]:
            value = assert_is_instance(trials[0][parameter], float)
            self.assertGreaterEqual(value, -1.0)
            self.assertLessEqual(value, 1.0)

        # Test can generate multiple trials
        trials = client.get_next_trials(max_trials=2)
        self.assertEqual(len(trials), 2)

        # Test respects fixed features
        with mock.patch(
            "ax.service.utils.with_db_settings_base"
            "._save_generation_strategy_to_db_if_possible"
        ) as mock_save:
            trials = client.get_next_trials(max_trials=1, fixed_parameters={"x1": 0.5})
        value = assert_is_instance(trials[3]["x1"], float)
        self.assertEqual(value, 0.5)

        # Check that GS is not saved to the DB.
        mock_save.assert_not_called()

    def test_get_next_trials_with_db(self) -> None:
        init_test_engine_and_session_factory(force_init=True)
        client = Client(storage_config=StorageConfig())

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
                RangeParameterConfig(name="x2", parameter_type="float", bounds=(-1, 1)),
            ],
            name="foo",
        )

        client.configure_optimization(objective="foo")
        client.configure_generation_strategy(
            initialization_budget=1, initialize_with_center=False
        )

        # Test can generate one trial
        trials = client.get_next_trials(max_trials=1)
        self.assertEqual(len(trials), 1)
        self.assertIn(0, trials)
        trial = client._experiment.trials[0]
        self.assertEqual(trial.generator_runs[0]._model_key, "Sobol")
        client.complete_trial(
            trial_index=trial.index, raw_data={"foo": (random.random(), 1.0)}
        )
        # Generate one more trial, so that GS transitions to BO.
        with mock.patch(
            "ax.service.utils.with_db_settings_base"
            "._save_generation_strategy_to_db_if_possible",
            wraps=_save_generation_strategy_to_db_if_possible,
        ) as mock_save:
            trials = client.get_next_trials(max_trials=1)
        # Check that GS was saved after generating the trials.
        self.assertEqual(client._generation_strategy.current_node_name, "MBM")
        mock_save.assert_called_once()

        # Check that loading the GS from the DB results in using BO.
        # This will only happen if the GS is saved in get_next_trials
        client2 = Client.load_from_database(
            client._experiment.name, storage_config=StorageConfig()
        )
        self.assertEqual(client2._generation_strategy.current_node_name, "MBM")
        trials = client2.get_next_trials(max_trials=1)
        self.assertEqual(len(trials), 1)
        self.assertIn(2, trials)
        trial = client2._experiment.trials[2]
        self.assertEqual(trial.generator_runs[0]._model_key, "BoTorch")

    def test_attach_data(self) -> None:
        client = Client()

        with self.assertRaisesRegex(AssertionError, "Experiment not set"):
            client.attach_data(trial_index=0, raw_data={"foo": 1.0})

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
            ],
            name="foo",
        )
        client.configure_optimization(objective="foo")

        # Vanilla case with no progression argument
        trial_index = [*client.get_next_trials(max_trials=1).keys()][0]
        client.attach_data(trial_index=trial_index, raw_data={"foo": 1.0})

        self.assertEqual(
            client._experiment.trials[trial_index].status,
            TrialStatus.RUNNING,
        )
        self.assertTrue(
            assert_is_instance(
                client._experiment.lookup_data(trial_indices=[trial_index]),
                MapData,
            ).map_df.equals(
                pd.DataFrame(
                    {
                        "trial_index": {0: 0},
                        "arm_name": {0: "0_0"},
                        "metric_name": {0: "foo"},
                        "mean": {0: 1.0},
                        "sem": {0: np.nan},
                        "step": {0: np.nan},
                    }
                )
            ),
        )

        # With progression argument
        client.attach_data(trial_index=0, raw_data={"foo": 2.0}, progression=10)

        self.assertEqual(
            client._experiment.trials[trial_index].status,
            TrialStatus.RUNNING,
        )
        self.assertTrue(
            assert_is_instance(
                client._experiment.lookup_data(trial_indices=[trial_index]),
                MapData,
            ).map_df.equals(
                pd.DataFrame(
                    {
                        "trial_index": {0: 0, 1: 0},
                        "arm_name": {0: "0_0", 1: "0_0"},
                        "metric_name": {0: "foo", 1: "foo"},
                        "mean": {0: 1.0, 1: 2.0},
                        "sem": {0: np.nan, 1: np.nan},
                        "step": {0: np.nan, 1: 10.0},
                    }
                )
            ),
        )

        # With extra metrics
        client.attach_data(
            trial_index=trial_index,
            raw_data={"foo": 1.0, "bar": 2.0},
        )
        self.assertEqual(
            client._experiment.trials[trial_index].status,
            TrialStatus.RUNNING,
        )
        self.assertTrue(
            assert_is_instance(
                client._experiment.lookup_data(trial_indices=[trial_index]),
                MapData,
            ).map_df.equals(
                pd.DataFrame(
                    {
                        "trial_index": {0: 0, 1: 0, 2: 0},
                        "arm_name": {0: "0_0", 1: "0_0", 2: "0_0"},
                        "metric_name": {0: "foo", 1: "foo", 2: "bar"},
                        "mean": {0: 2.0, 1: 1.0, 2: 2.0},
                        "sem": {0: np.nan, 1: np.nan, 2: np.nan},
                        "step": {0: 10.0, 1: np.nan, 2: np.nan},
                    }
                )
            ),
        )

    def test_complete_trial(self) -> None:
        client = Client()

        with self.assertRaisesRegex(AssertionError, "Experiment not set"):
            client.complete_trial(trial_index=0, raw_data={"foo": 1.0})

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
            ],
            name="foo",
        )
        client.configure_optimization(objective="foo", outcome_constraints=["bar >= 0"])

        # Vanilla case with no progression argument
        trial_index = [*client.get_next_trials(max_trials=1).keys()][0]
        client.complete_trial(
            trial_index=trial_index, raw_data={"foo": 1.0, "bar": 2.0}
        )

        self.assertEqual(
            client._experiment.trials[trial_index].status,
            TrialStatus.COMPLETED,
        )
        self.assertTrue(
            assert_is_instance(
                client._experiment.lookup_data(trial_indices=[trial_index]),
                MapData,
            ).map_df.equals(
                pd.DataFrame(
                    {
                        "trial_index": {0: 0, 1: 0},
                        "arm_name": {0: "0_0", 1: "0_0"},
                        "metric_name": {0: "foo", 1: "bar"},
                        "mean": {0: 1.0, 1: 2.0},
                        "sem": {0: np.nan, 1: np.nan},
                        "step": {0: np.nan, 1: np.nan},
                    }
                )
            ),
        )

        # With progression argument
        trial_index = [*client.get_next_trials(max_trials=1).keys()][0]
        client.complete_trial(
            trial_index=trial_index, raw_data={"foo": 1.0, "bar": 2.0}, progression=10
        )

        self.assertEqual(
            client._experiment.trials[trial_index].status,
            TrialStatus.COMPLETED,
        )

        self.assertTrue(
            assert_is_instance(
                client._experiment.lookup_data(trial_indices=[trial_index]),
                MapData,
            ).map_df.equals(
                pd.DataFrame(
                    {
                        "trial_index": {0: 1, 1: 1},
                        "arm_name": {0: "1_0", 1: "1_0"},
                        "metric_name": {0: "foo", 1: "bar"},
                        "mean": {0: 1.0, 1: 2.0},
                        "sem": {0: np.nan, 1: np.nan},
                        "step": {0: 10.0, 1: 10.0},
                    }
                )
            ),
        )

        # With missing metrics
        trial_index = [*client.get_next_trials(max_trials=1).keys()][0]
        client.complete_trial(trial_index=trial_index, raw_data={"foo": 1.0})

        self.assertEqual(
            client._experiment.trials[trial_index].status,
            TrialStatus.FAILED,
        )
        self.assertTrue(
            assert_is_instance(
                client._experiment.lookup_data(trial_indices=[trial_index]),
                MapData,
            ).map_df.equals(
                pd.DataFrame(
                    {
                        "trial_index": {0: 2},
                        "arm_name": {0: "2_0"},
                        "metric_name": {0: "foo"},
                        "mean": {0: 1.0},
                        "sem": {0: np.nan},
                        "step": {0: np.nan},
                    }
                )
            ),
        )

    def test_attach_trial(self) -> None:
        client = Client()

        with self.assertRaisesRegex(AssertionError, "Experiment not set"):
            client.attach_trial(parameters={"x1": 0.5})

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
            ],
            name="foo",
        )
        client.configure_optimization(objective="foo")

        trial_index = client.attach_trial(parameters={"x1": 0.5}, arm_name="bar")
        trial = assert_is_instance(client._experiment.trials[trial_index], Trial)
        self.assertEqual(none_throws(trial.arm).parameters, {"x1": 0.5})
        self.assertEqual(none_throws(trial.arm).name, "bar")
        self.assertEqual(trial.status, TrialStatus.RUNNING)

    def test_attach_baseline(self) -> None:
        client = Client()

        with self.assertRaisesRegex(AssertionError, "Experiment not set"):
            client.attach_baseline(parameters={"x1": 0.5})

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
            ],
            name="foo",
        )
        client.configure_optimization(objective="foo")

        trial_index = client.attach_baseline(parameters={"x1": 0.5})
        trial = assert_is_instance(client._experiment.trials[trial_index], Trial)
        self.assertEqual(none_throws(trial.arm).parameters, {"x1": 0.5})
        self.assertEqual(none_throws(trial.arm).name, "baseline")
        self.assertEqual(trial.status, TrialStatus.RUNNING)

        self.assertEqual(client._experiment.status_quo, trial.arm)

    def test_mark_trial_failed(self) -> None:
        client = Client()

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
            ],
            name="foo",
        )
        client.configure_optimization(objective="foo")

        trial_index = [*client.get_next_trials(max_trials=1).keys()][0]
        client.mark_trial_failed(
            trial_index=trial_index, failed_reason="testing the optional parameter"
        )
        self.assertEqual(
            client._experiment.trials[trial_index].status,
            TrialStatus.FAILED,
        )
        self.assertEqual(
            client._experiment.trials[trial_index]._failed_reason,
            "testing the optional parameter",
        )

    def test_mark_trial_abandoned(self) -> None:
        client = Client()

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
            ],
            name="foo",
        )
        client.configure_optimization(objective="foo")

        trial_index = [*client.get_next_trials(max_trials=1).keys()][0]
        client.mark_trial_abandoned(trial_index=trial_index)
        self.assertEqual(
            client._experiment.trials[trial_index].status,
            TrialStatus.ABANDONED,
        )

    def test_mark_trial_early_stopped(self) -> None:
        client = Client()

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
            ],
            name="foo",
        )
        client.configure_optimization(objective="foo")

        trial_index = [*client.get_next_trials(max_trials=1).keys()][0]

        with self.assertRaisesRegex(
            UnsupportedError, "Cannot mark trial early stopped"
        ):
            client.mark_trial_early_stopped(trial_index=trial_index)

        client.attach_data(
            trial_index=trial_index, raw_data={"foo": 0.0}, progression=1
        )
        client.mark_trial_early_stopped(trial_index=trial_index)
        self.assertEqual(
            client._experiment.trials[trial_index].status,
            TrialStatus.EARLY_STOPPED,
        )
        self.assertTrue(
            assert_is_instance(
                client._experiment.lookup_data(trial_indices=[trial_index]),
                MapData,
            ).map_df.equals(
                pd.DataFrame(
                    {
                        "trial_index": {0: 0},
                        "arm_name": {0: "0_0"},
                        "metric_name": {0: "foo"},
                        "mean": {0: 0.0},
                        "sem": {0: np.nan},
                        "step": {0: 1.0},
                    }
                )
            ),
        )

    def test_should_stop_trial_early(self) -> None:
        client = Client()

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
            ],
            name="foo",
        )
        client.configure_optimization(objective="foo")

        client.set_early_stopping_strategy(
            early_stopping_strategy=PercentileEarlyStoppingStrategy(
                metric_names=["foo"]
            )
        )

        client.get_next_trials(max_trials=1)
        self.assertFalse(client.should_stop_trial_early(trial_index=0))

    def test_run_trials(self) -> None:
        client = Client()

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
            ],
            name="foo",
        )
        client.configure_optimization(objective="foo")
        client.configure_metrics(metrics=[DummyMetric(name="foo")])
        client.configure_runner(runner=DummyRunner())

        client.run_trials(max_trials=4)

        self.assertEqual(len(client._experiment.trials), 4)
        self.assertEqual(
            [
                trial.index
                for trial in client._experiment.trials_by_status[TrialStatus.COMPLETED]
            ],
            [0, 1, 2, 3],
        )

        self.assertTrue(
            assert_is_instance(
                client._experiment.lookup_data(),
                MapData,
            ).map_df.equals(
                pd.DataFrame(
                    {
                        "trial_index": {0: 0, 1: 1, 2: 2, 3: 3},
                        "arm_name": {0: "0_0", 1: "1_0", 2: "2_0", 3: "3_0"},
                        "metric_name": {0: "foo", 1: "foo", 2: "foo", 3: "foo"},
                        "mean": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                        "sem": {0: np.nan, 1: np.nan, 2: np.nan, 3: np.nan},
                        "step": {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0},
                    }
                )
            ),
        )

    def test_get_next_trials_then_run_trials(self) -> None:
        client = Client()

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
            ],
            name="foo",
        )
        client.configure_optimization(objective="foo")
        client.configure_metrics(metrics=[DummyMetric(name="foo")])
        client.configure_runner(runner=DummyRunner())

        # First use Client in ask-tell
        # Complete two trials
        for index, _parameters in client.get_next_trials(max_trials=2).items():
            client.complete_trial(trial_index=index, raw_data={"foo": 1.0})

        # Leave one trial RUNNING
        _ = client.get_next_trials(max_trials=1)

        self.assertEqual(
            len(client._experiment.trials_by_status[TrialStatus.COMPLETED]),
            2,
        )
        self.assertEqual(
            len(client._experiment.trials_by_status[TrialStatus.RUNNING]),
            1,
        )

        # Configure runners and Metrics Run another two trials
        client.configure_metrics(metrics=[DummyMetric(name="foo")])
        client.configure_runner(runner=DummyRunner())
        client.run_trials(max_trials=2)

        # All trials should be COMPLETED
        self.assertEqual(
            len(client._experiment.trials_by_status[TrialStatus.COMPLETED]),
            5,
        )

    def test_summarize(self) -> None:
        client = Client()

        client.configure_experiment(
            name="test_experiment",
            parameters=[
                RangeParameterConfig(
                    name="x1",
                    parameter_type="float",
                    bounds=(0, 1),
                ),
                RangeParameterConfig(
                    name="x2",
                    parameter_type="float",
                    bounds=(0, 1),
                ),
            ],
        )
        client.configure_optimization(objective="foo, bar")
        # No GS, no data.
        summary_df = client.summarize()
        self.assertTrue(summary_df.empty)

        # Add manual trial.
        index = client.attach_trial(
            parameters={"x1": 0.5, "x2": 0.5}, arm_name="manual"
        )
        client.attach_data(trial_index=index, raw_data={"foo": 0.0, "bar": 0.5})
        summary_df = client.summarize()
        expected_columns = {
            "trial_index",
            "arm_name",
            "trial_status",
            "foo",
            "bar",
            "x1",
            "x2",
        }
        self.assertEqual(set(summary_df.columns), expected_columns)

        # Get two trials and fail one, giving us a ragged structure
        client.get_next_trials(max_trials=2)
        client.complete_trial(trial_index=1, raw_data={"foo": 1.0, "bar": 2.0})
        client.mark_trial_failed(trial_index=2)

        summary_df = client.summarize()
        expected_columns.add("generation_node")
        self.assertEqual(set(summary_df.columns), expected_columns)

        trial_0_parameters = none_throws(
            assert_is_instance(client._experiment.trials[0], Trial).arm
        ).parameters
        trial_1_parameters = none_throws(
            assert_is_instance(client._experiment.trials[1], Trial).arm
        ).parameters
        trial_2_parameters = none_throws(
            assert_is_instance(client._experiment.trials[2], Trial).arm
        ).parameters
        expected = pd.DataFrame(
            {
                "trial_index": {0: 0, 1: 1, 2: 2},
                "arm_name": {0: "manual", 1: "1_0", 2: "2_0"},
                "trial_status": {0: "RUNNING", 1: "COMPLETED", 2: "FAILED"},
                "generation_node": {0: None, 1: "CenterOfSearchSpace", 2: "Sobol"},
                "foo": {0: 0.0, 1: 1.0, 2: np.nan},  # NaN because trial 2 failed
                "bar": {0: 0.5, 1: 2.0, 2: np.nan},
                "x1": {
                    0: trial_0_parameters["x1"],
                    1: trial_1_parameters["x1"],
                    2: trial_2_parameters["x1"],
                },
                "x2": {
                    0: trial_0_parameters["x2"],
                    1: trial_1_parameters["x2"],
                    2: trial_2_parameters["x2"],
                },
            }
        )
        pd.testing.assert_frame_equal(summary_df, expected)

        # Test with trial_indices parameter
        # Only include trials 0 and 1
        summary_df_filtered = client.summarize(trial_indices=[0, 1])
        expected_filtered = pd.DataFrame(
            {
                "trial_index": {0: 0, 1: 1},
                "arm_name": {0: "manual", 1: "1_0"},
                "trial_status": {0: "RUNNING", 1: "COMPLETED"},
                "generation_node": {0: None, 1: "CenterOfSearchSpace"},
                "foo": {0: 0.0, 1: 1.0},
                "bar": {0: 0.5, 1: 2.0},
                "x1": {
                    0: trial_0_parameters["x1"],
                    1: trial_1_parameters["x1"],
                },
                "x2": {
                    0: trial_0_parameters["x2"],
                    1: trial_1_parameters["x2"],
                },
            }
        )
        pd.testing.assert_frame_equal(summary_df_filtered, expected_filtered)

        # Test with only one trial index
        summary_df_single = client.summarize(trial_indices=[1])
        expected_single = pd.DataFrame(
            {
                "trial_index": {0: 1},
                "arm_name": {0: "1_0"},
                "trial_status": {0: "COMPLETED"},
                "generation_node": {0: "CenterOfSearchSpace"},
                "foo": {0: 1.0},
                "bar": {0: 2.0},
                "x1": {0: trial_1_parameters["x1"]},
                "x2": {0: trial_1_parameters["x2"]},
            }
        )
        pd.testing.assert_frame_equal(summary_df_single, expected_single)

        # Test with trial_status parameter
        summary_df_completed = client.summarize(trial_statuses=["completed"])
        expected_completed = pd.DataFrame(
            {
                "trial_index": {0: 1},
                "arm_name": {0: "1_0"},
                "trial_status": {0: "COMPLETED"},
                "generation_node": {0: "CenterOfSearchSpace"},
                "foo": {0: 1.0},
                "bar": {0: 2.0},
                "x1": {0: trial_1_parameters["x1"]},
                "x2": {0: trial_1_parameters["x2"]},
            }
        )
        pd.testing.assert_frame_equal(summary_df_completed, expected_completed)

        # Test with trial_status parameter for running trials
        summary_df_running = client.summarize(trial_statuses=["running"])
        expected_running = pd.DataFrame(
            {
                "trial_index": {0: 0},
                "arm_name": {0: "manual"},
                "trial_status": {0: "RUNNING"},
                "foo": {0: 0.0},
                "bar": {0: 0.5},
                "x1": {0: trial_0_parameters["x1"]},
                "x2": {0: trial_0_parameters["x2"]},
            }
        )

        assert summary_df_running.equals(expected_running)

        # Test with multiple trial_status values
        summary_df_multi_status = client.summarize(
            trial_statuses=["completed", "running"]
        )
        expected_multi_status = pd.DataFrame(
            {
                "trial_index": {0: 0, 1: 1},
                "arm_name": {0: "manual", 1: "1_0"},
                "trial_status": {0: "RUNNING", 1: "COMPLETED"},
                "generation_node": {0: None, 1: "CenterOfSearchSpace"},
                "foo": {0: 0.0, 1: 1.0},
                "bar": {0: 0.5, 1: 2.0},
                "x1": {
                    0: trial_0_parameters["x1"],
                    1: trial_1_parameters["x1"],
                },
                "x2": {
                    0: trial_0_parameters["x2"],
                    1: trial_1_parameters["x2"],
                },
            }
        )
        self.assertTrue(summary_df_multi_status.equals(expected_multi_status))

    def test_compute_analyses(self) -> None:
        client = Client()

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
            ],
            name="foo",
        )
        client.configure_optimization(objective="foo")
        client.configure_generation_strategy()

        with self.assertLogs(logger="ax.analysis", level="ERROR") as lg:
            analysis = ParallelCoordinatesPlot()
            cards = client.compute_analyses(analyses=[analysis])

            self.assertEqual(len(cards), 1)
            self.assertEqual(cards[0].name, "ParallelCoordinatesPlot")
            self.assertEqual(cards[0].title, "ParallelCoordinatesPlot Error")
            self.assertEqual(
                cards[0].subtitle,
                "ValueError encountered while computing ParallelCoordinatesPlot.",
            )
            self.assertIn("Traceback", assert_is_instance(cards[0], AnalysisCard).blob)
            self.assertTrue(
                any(("No data found for metric") in msg for msg in lg.output)
            )

        for trial_index, _ in client.get_next_trials(max_trials=1).items():
            client.complete_trial(trial_index=trial_index, raw_data={"foo": 1.0})

        cards = client.compute_analyses(analyses=[ParallelCoordinatesPlot()])

        self.assertEqual(len(cards), 1)
        self.assertEqual(cards[0].name, "ParallelCoordinatesPlot")

    @mock_botorch_optimize
    def test_get_best_parameterization(self) -> None:
        client = Client()

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
            ],
            name="foo",
        )

        with self.assertRaisesRegex(
            UnsupportedError, "No optimization config has been set"
        ):
            client.get_best_parameterization()

        client.configure_optimization(objective="foo")
        # Set initialization_budget=3 so we can reach a predictive GenerationNode
        # quickly
        client.configure_generation_strategy(initialization_budget=3)

        with self.assertRaisesRegex(UnsupportedError, "No trials have been run yet"):
            client.get_best_parameterization()

        for _ in range(3):
            for index, parameters in client.get_next_trials(max_trials=1).items():
                client.complete_trial(
                    trial_index=index,
                    raw_data={
                        "foo": assert_is_instance(parameters["x1"], float) ** 2,
                    },
                )

        parameters, prediction, index, name = client.get_best_parameterization()
        self.assertIn(
            name,
            [
                none_throws(assert_is_instance(trial, Trial).arm).name
                for trial in client._experiment.trials.values()
            ],
        )
        self.assertTrue(
            client._experiment.search_space.check_membership(
                parameterization=parameters
            )
        )
        self.assertEqual({*prediction.keys()}, {"foo"})

        # Run a non-Sobol trial
        for index, parameters in client.get_next_trials(max_trials=1).items():
            client.complete_trial(
                trial_index=index,
                raw_data={
                    "foo": assert_is_instance(parameters["x1"], float) ** 2,
                },
            )
        parameters, prediction, index, name = client.get_best_parameterization()
        self.assertIn(
            name,
            [
                none_throws(assert_is_instance(trial, Trial).arm).name
                for trial in client._experiment.trials.values()
            ],
        )
        self.assertTrue(
            client._experiment.search_space.check_membership(
                parameterization=parameters
            )
        )
        self.assertEqual({*prediction.keys()}, {"foo"})

        # Try calling after setting OptimizationConfig to MOO problem
        client.configure_optimization("foo, bar")
        with self.assertRaisesRegex(
            UnsupportedError, "Please call get_pareto_frontier"
        ):
            client.get_best_parameterization()

    @mock_botorch_optimize
    def test_get_pareto_frontier(self) -> None:
        client = Client()

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
            ],
            name="foo",
        )

        with self.assertRaisesRegex(
            UnsupportedError, "No optimization config has been set"
        ):
            client.get_pareto_frontier()

        client.configure_optimization(objective="foo, bar")
        # Set initialization_budget=3 so we can reach a predictive GenerationNode
        # quickly
        client.configure_generation_strategy(initialization_budget=3)

        with self.assertRaisesRegex(UnsupportedError, "No trials have been run yet"):
            client.get_pareto_frontier()

        for _ in range(3):
            for index, parameters in client.get_next_trials(max_trials=1).items():
                client.complete_trial(
                    trial_index=index,
                    raw_data={
                        "foo": assert_is_instance(parameters["x1"], float) ** 2,
                        "bar": 0.0,
                    },
                )

        frontier = client.get_pareto_frontier(False)
        for parameters, prediction, index, name in frontier:
            self.assertEqual(
                none_throws(
                    assert_is_instance(client._experiment.trials[index], Trial).arm
                ).name,
                name,
            )
            self.assertIn(
                name,
                [
                    none_throws(assert_is_instance(trial, Trial).arm).name
                    for trial in client._experiment.trials.values()
                ],
            )
            self.assertTrue(
                client._experiment.search_space.check_membership(
                    parameterization=parameters
                )
            )
            self.assertEqual({*prediction.keys()}, {"foo", "bar"})

        # Run a non-Sobol trial
        for index, parameters in client.get_next_trials(max_trials=1).items():
            client.complete_trial(
                trial_index=index,
                raw_data={
                    "foo": assert_is_instance(parameters["x1"], float) ** 2,
                    "bar": 0.0,
                },
            )
        frontier = client.get_pareto_frontier()
        for parameters, prediction, index, name in frontier:
            self.assertEqual(
                none_throws(
                    assert_is_instance(client._experiment.trials[index], Trial).arm
                ).name,
                name,
            )
            self.assertIn(
                name,
                [
                    none_throws(assert_is_instance(trial, Trial).arm).name
                    for trial in client._experiment.trials.values()
                ],
            )
            self.assertTrue(
                client._experiment.search_space.check_membership(
                    parameterization=parameters
                )
            )
            self.assertEqual({*prediction.keys()}, {"foo", "bar"})

        # Try calling after setting OptimizationConfig to single objective problem
        client.configure_optimization("foo")
        with self.assertRaisesRegex(
            UnsupportedError, "Please call get_best_parameterization"
        ):
            client.get_pareto_frontier()

    @mock_botorch_optimize
    def test_predict(self) -> None:
        client = Client()

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
            ],
            name="foo",
        )
        client.configure_optimization(objective="foo", outcome_constraints=["bar >= 0"])
        # Set initialization_budget=3 so we can reach a predictive GenerationNode
        # quickly
        client.configure_generation_strategy(initialization_budget=3)

        with self.assertRaisesRegex(ValueError, "but search space has parameters"):
            client.predict(points=[{"x0": 0}])

        with self.assertRaisesRegex(UnsupportedError, "not predictive"):
            client.predict(points=[{"x1": 0}])

        client.configure_metrics(metrics=[DummyMetric(name="baz")])
        for _ in range(4):
            for index, parameters in client.get_next_trials(max_trials=1).items():
                client.complete_trial(
                    trial_index=index,
                    raw_data={
                        "foo": assert_is_instance(parameters["x1"], float) ** 2,
                        "bar": 0.0,
                    },
                )

        # Check we've predicted something for foo and bar but not baz (which is a
        # tracking metric)
        (prediction,) = client.predict(points=[{"x1": 0.5}])
        self.assertEqual(set(prediction), {"foo", "bar"})
        # Check that we're returning SEM not variance.
        pred_mean = {"foo": [0.25], "bar": [0.0]}
        pred_cov = {
            "foo": {"foo": [4.0], "bar": [9.0]},
            "bar": {"foo": [9.0], "bar": [16.0]},
        }
        with mock.patch(
            "ax.adapter.torch.TorchAdapter.predict", return_value=(pred_mean, pred_cov)
        ) as mock_predict:
            (prediction,) = client.predict(points=[{"x1": 0.5}])
        mock_predict.assert_called_once()
        self.assertEqual(prediction, {"foo": (0.25, 2.0), "bar": (0.0, 4.0)})

    def test_json_storage(self) -> None:
        client = Client()

        # Experiment with relatively complicated search space
        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
                RangeParameterConfig(name="x2", parameter_type="int", bounds=(-1, 1)),
                ChoiceParameterConfig(
                    name="x3",
                    parameter_type="str",
                    values=["a", "b"],
                ),
                ChoiceParameterConfig(
                    name="x4",
                    parameter_type="int",
                    values=[1, 2, 3],
                    is_ordered=True,
                ),
                ChoiceParameterConfig(name="x5", parameter_type="int", values=[1]),
            ],
            name="foo",
        )

        # Relatively complicated optimization config
        client.configure_optimization(
            objective="foo + 2 * bar", outcome_constraints=["baz >= 0"]
        )

        # Specified generation strategy
        client.configure_generation_strategy(
            initialization_budget=2,
        )

        # Use the Client a bit
        _ = client.get_next_trials(max_trials=2)

        snapshot = client._to_json_snapshot()
        other_client = Client._from_json_snapshot(snapshot=snapshot)

        self.assertEqual(client._experiment, other_client._experiment)
        # Don't check for deep equality of GenerationStrategy since the other gs will
        # not have all its attributes initialized, but ensure they have the same repr
        self.assertEqual(
            str(client._generation_strategy), str(other_client._generation_strategy)
        )

    def test_sql_storage(self) -> None:
        init_test_engine_and_session_factory(force_init=True)
        client = Client(storage_config=StorageConfig())

        # Experiment with relatively complicated search space
        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(-1, 1)),
                RangeParameterConfig(name="x2", parameter_type="int", bounds=(-1, 1)),
                ChoiceParameterConfig(
                    name="x3",
                    parameter_type="str",
                    values=["a", "b"],
                ),
                ChoiceParameterConfig(
                    name="x4",
                    parameter_type="int",
                    values=[1, 2, 3],
                    is_ordered=True,
                ),
                ChoiceParameterConfig(name="x5", parameter_type="int", values=[1]),
            ],
            name="unique_test_experiment",
        )

        # Relatively complicated optimization config
        client.configure_optimization(
            objective="foo + 2 * bar", outcome_constraints=["baz >= 0"]
        )

        # Specified generation strategy
        client.configure_generation_strategy(
            initialization_budget=3,
        )

        other_client = Client.load_from_database(
            experiment_name="unique_test_experiment", storage_config=StorageConfig()
        )

        self.assertEqual(client._experiment, other_client._experiment)
        # Don't check for deep equality of GenerationStrategy since the other gs will
        # not have all its attributes initialized, but ensure they have the same repr
        self.assertEqual(
            str(client._generation_strategy), str(other_client._generation_strategy)
        )

    def test_overwrite_metric(self) -> None:
        client = Client()

        client.configure_experiment(
            parameters=[
                RangeParameterConfig(name="x1", parameter_type="float", bounds=(0, 1))
            ],
            name="foo_exp",
        )

        # Start with single-objective + outcome constraint
        client.configure_optimization(objective="foo", outcome_constraints=["bar >= 0"])

        # test 1: Replace the single-objective metric
        foo_metric = DummyMetric(name="foo")
        client.configure_metrics(metrics=[foo_metric])
        objective = none_throws(client._experiment.optimization_config).objective
        self.assertIsInstance(objective, Objective)
        self.assertIs(objective.metric, foo_metric)

        # test 2: Replace the outcome-constraint metric
        bar_metric = DummyMetric(name="bar")
        client.configure_metrics(metrics=[bar_metric])
        oc = none_throws(client._experiment.optimization_config).outcome_constraints[0]
        self.assertIs(oc.metric, bar_metric)

        # test 3: Add a tracking metric, then replace it by name
        baz_metric_1 = DummyMetric(name="baz")
        client.configure_metrics(metrics=[baz_metric_1])
        self.assertIn("baz", client._experiment._tracking_metrics)
        self.assertIs(client._experiment._tracking_metrics["baz"], baz_metric_1)

        baz_metric_2 = DummyMetric(name="baz")
        client.configure_metrics(metrics=[baz_metric_2])
        self.assertIs(client._experiment._tracking_metrics["baz"], baz_metric_2)

        # test 4: Metric name not present anywhere, should be added as tracking + warn
        quux_metric = DummyMetric(name="quux")
        with self.assertLogs(logger="ax.api.client", level="WARNING") as logs:
            client.configure_metrics(metrics=[quux_metric])
        self.assertIn("quux", client._experiment._tracking_metrics)
        self.assertIs(client._experiment._tracking_metrics["quux"], quux_metric)
        self.assertTrue(
            any("added as tracking metric" in msg for msg in logs.output),
            "Expected a warning that the metric was added as a tracking metric.",
        )

        # test 5: Replace inside a MultiObjective
        client.configure_optimization(objective="foo, qux")
        qux_metric_moo = DummyMetric(name="qux")
        client.configure_metrics(metrics=[qux_metric_moo])
        moo = assert_is_instance(
            none_throws(client._experiment.optimization_config).objective,
            MultiObjective,
        )
        self.assertIn(qux_metric_moo, moo.metrics)

        # test 6: Replace inside a ScalarizedObjective
        client.configure_optimization(objective="foo + qux")
        qux_metric_scalar = DummyMetric(name="qux")
        client.configure_metrics(metrics=[qux_metric_scalar])
        scalar = assert_is_instance(
            none_throws(client._experiment.optimization_config).objective,
            ScalarizedObjective,
        )
        self.assertIn(qux_metric_scalar, scalar.metrics)


class DummyRunner(IRunner):
    @override
    def run_trial(
        self, trial_index: int, parameterization: TParameterization
    ) -> dict[str, Any]:
        return {}

    @override
    def poll_trial(
        self, trial_index: int, trial_metadata: Mapping[str, Any]
    ) -> TrialStatus:
        return TrialStatus.COMPLETED

    @override
    def stop_trial(
        self, trial_index: int, trial_metadata: Mapping[str, Any]
    ) -> dict[str, Any]: ...


class DummyMetric(IMetric):
    def fetch(
        self,
        trial_index: int,
        trial_metadata: Mapping[str, Any],
    ) -> tuple[int, float | tuple[float, float]]:
        return 0, 0.0



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import os


def is_running_in_papermill() -> bool:
    return os.environ.get("RUNNING_IN_PAPERMILL") == "True"



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field

from ax.benchmark.benchmark_metric import BenchmarkMapMetric, BenchmarkMetric

from ax.benchmark.benchmark_step_runtime_function import TBenchmarkStepRuntimeFunction
from ax.benchmark.benchmark_test_function import BenchmarkTestFunction
from ax.core.auxiliary import AuxiliaryExperiment, AuxiliaryExperimentPurpose

from ax.core.objective import MultiObjective, Objective
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    ObjectiveThreshold,
    OptimizationConfig,
)
from ax.core.outcome_constraint import OutcomeConstraint
from ax.core.parameter import ChoiceParameter, ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.core.types import ComparisonOp, TParamValue
from ax.exceptions.core import UserInputError
from ax.utils.common.base import Base


@dataclass(kw_only=True, repr=True)
class BenchmarkProblem(Base):
    """
    Problem against which different methods can be benchmarked.

    Defines how data is generated, the objective (via the OptimizationConfig),
    and the SearchSpace.

    Args:
        name: Can be generated programmatically with `_get_name`.
        optimization_config: Defines the objective of optimization. Metrics must
            be `BenchmarkMetric`s.
        num_trials: Number of optimization iterations to run. BatchTrials count
            as one trial.
        optimal_value: The best ground-truth objective value, used for scoring
            optimization results on a scale from 0 to 100, where achieving the
            `optimal_value` receives a score of 100. The `optimal_value` should
            be a hypervolume for multi-objective problems. If the best value is
            not known, it is conventional to set it to a value that is almost
            certainly better than the best value, so that a benchmark's score
            will not exceed 100%.
        baseline_value: Similar to `optimal_value`, but a not-so-good value
            which benchmarks are expected to do better than. A baseline value
            can be derived using the function
            `compute_baseline_value_from_sobol`, which takes the best of five
            quasi-random Sobol trials.
        search_space: The search space.
        test_function: A `BenchmarkTestFunction`, which will generate noiseless
            data. This will be used by a `BenchmarkRunner`.
        noise_std: Describes how noise is added to the output of the
            `test_function`. If a float, IID random normal noise with that
            standard deviation is added. A list of floats, or a dict whose keys
            match `test_functions.outcome_names`, sets different noise
            standard deviations for the different outcomes produced by the
            `test_function`. This will be used by a `BenchmarkRunner`.
        report_inference_value_as_trace: Whether the ``optimization_trace`` on a
            ``BenchmarkResult`` should use the ``oracle_trace`` (if False,
            default) or the ``inference_trace``. See ``BenchmarkResult`` for
            more information. Currently, this is only supported for
            single-objective problems.
        step_runtime_function: Optionally, a function that takes in ``params``
            (typically dictionaries mapping strings to ``TParamValue``s) and
            returns the runtime of an step. If ``step_runtime_function`` is
            left as ``None``, each step will take one simulated second.  (When
            data is not time-series, the whole trial consists of one step.)
        auxiliary_experiments_by_purpose: A mapping from experiment purpose to
            a list of auxiliary experiments.
    """

    name: str
    optimization_config: OptimizationConfig
    num_trials: int
    test_function: BenchmarkTestFunction
    noise_std: float | Sequence[float] | Mapping[str, float] = 0.0
    optimal_value: float
    baseline_value: float
    search_space: SearchSpace = field(repr=False)
    report_inference_value_as_trace: bool = False
    step_runtime_function: TBenchmarkStepRuntimeFunction | None = None
    target_fidelity_and_task: Mapping[str, TParamValue] = field(default_factory=dict)
    status_quo_params: Mapping[str, TParamValue] | None = None
    auxiliary_experiments_by_purpose: (
        dict[AuxiliaryExperimentPurpose, list[AuxiliaryExperiment]] | None
    ) = None

    def __post_init__(self) -> None:
        # Validate inputs
        if self.report_inference_value_as_trace and self.is_moo:
            raise NotImplementedError(
                "Inference trace is not supported for MOO. Please set "
                "`report_inference_value_as_trace` to False."
            )

        # Validate that the optimal value is actually better than the baseline
        # value
        # If MOO, values represent hypervolumes
        if isinstance(self.optimization_config, MultiObjectiveOptimizationConfig):
            if self.baseline_value >= self.optimal_value:
                raise ValueError(
                    "The baseline value must be strictly less than the optimal "
                    "value for MOO problems. (These represent hypervolumes.)"
                )
        elif self.optimization_config.objective.minimize:
            if self.baseline_value <= self.optimal_value:
                raise ValueError(
                    "The baseline value must be strictly greater than the optimal "
                    "value for minimization problems."
                )
        elif self.baseline_value >= self.optimal_value:
            raise ValueError(
                "The baseline value must be strictly less than the optimal "
                "value for maximization problems."
            )

        # Validate that names on optimization config are contained in names on
        # test function
        objective = self.optimization_config.objective
        if isinstance(objective, MultiObjective):
            objective_names = {obj.metric.name for obj in objective.objectives}
        else:
            objective_names = {objective.metric.name}

        test_function_names = set(self.test_function.outcome_names)
        missing = objective_names - test_function_names
        if len(missing) > 0:
            raise ValueError(
                "The following objectives are defined on "
                "`optimization_config` but not included in "
                f"`runner.test_function.outcome_names`: {missing}."
            )

        constraints = self.optimization_config.outcome_constraints
        constraint_names = {c.metric.name for c in constraints}
        missing = constraint_names - test_function_names
        if len(missing) > 0:
            raise ValueError(
                "The following constraints are defined on "
                "`optimization_config` but not included in "
                f"`runner.test_function.outcome_names`: {missing}."
            )
        if any(c.relative for c in constraints) and self.status_quo_params is None:
            raise ValueError(
                "Relative constraints require specifying status_quo_params."
            )

        self.target_fidelity_and_task = {
            p.name: p.target_value
            for p in self.search_space.parameters.values()
            if (isinstance(p, ChoiceParameter) and p.is_task) or p.is_fidelity
        }
        if (
            self.status_quo_params is not None
            and not self.search_space.check_membership(
                parameterization=self.status_quo_params
            )
        ):
            raise UserInputError("Status quo parameters are not in the search space.")

    @property
    def is_moo(self) -> bool:
        """Whether the problem is multi-objective."""
        return isinstance(self.optimization_config, MultiObjectiveOptimizationConfig)


def _get_constraints(
    constraint_names: Sequence[str],
    observe_noise_sd: bool,
    use_map_metric: bool = False,
) -> list[OutcomeConstraint]:
    """
    Create a list of ``OutcomeConstraint``s.

    Args:
        constraint_names: Names of the constraints. One constraint will be
            created for each.
        observe_noise_sd: Whether the standard deviation of the observation
            noise is observed, for each constraint. This doesn't handle the case
            where only some of the outcomes have noise levels observed.
        use_map_metric: Whether to use a ``BenchmarkMapMetric``.


    """
    metric_cls = BenchmarkMapMetric if use_map_metric else BenchmarkMetric
    outcome_constraints = [
        OutcomeConstraint(
            metric=metric_cls(
                name=name,
                lower_is_better=False,  # positive slack = feasible
                observe_noise_sd=observe_noise_sd,
            ),
            op=ComparisonOp.GEQ,
            bound=0.0,
            relative=False,
        )
        for name in constraint_names
    ]
    return outcome_constraints


def get_soo_opt_config(
    *,
    outcome_names: Sequence[str],
    lower_is_better: bool = True,
    observe_noise_sd: bool = False,
    use_map_metric: bool = False,
) -> OptimizationConfig:
    """
    Create a single-objective ``OptimizationConfig``, potentially with
    constraints.

    Args:
        outcome_names: Names of the outcomes. If ``outcome_names`` has more than
            one element, constraints will be created. The first element of
            ``outcome_names`` will be the name of the ``BenchmarkMetric`` on
            the objective, and others (if pressent) will be for constraints.
        lower_is_better: Whether the objective is a minimization problem. This
            only affects objectives, not constraints; for constraints, higher is
            better (feasible).
        observe_noise_sd: Whether the standard deviation of the observation
            noise is observed. Applies to all objective and constraints.
        use_map_metric: Whether to use a ``BenchmarkMapMetric``.
    """
    metric_cls = BenchmarkMapMetric if use_map_metric else BenchmarkMetric
    objective = Objective(
        metric=metric_cls(
            name=outcome_names[0],
            lower_is_better=lower_is_better,
            observe_noise_sd=observe_noise_sd,
        ),
        minimize=lower_is_better,
    )

    outcome_constraints = _get_constraints(
        constraint_names=outcome_names[1:],
        observe_noise_sd=observe_noise_sd,
        use_map_metric=use_map_metric,
    )

    return OptimizationConfig(
        objective=objective, outcome_constraints=outcome_constraints
    )


def get_moo_opt_config(
    *,
    outcome_names: Sequence[str],
    ref_point: Sequence[float],
    num_constraints: int = 0,
    lower_is_better: bool = True,
    observe_noise_sd: bool = False,
    use_map_metric: bool = False,
) -> MultiObjectiveOptimizationConfig:
    """
    Create a ``MultiObjectiveOptimizationConfig``, potentially with constraints.

    Args:
        outcome_names: Names of the outcomes. If ``num_constraints`` is greater
            than zero, the last ``num_constraints`` elements of
            ``outcome_names`` will become the names of ``BenchmarkMetric``s on
            constraints, and the others will correspond to the objectives.
        ref_point: Objective thresholds for the objective metrics. Note:
            Although this method requires providing a threshold for each
            objective, this is not required in general and could be enabled for
            this method.
        num_constraints: Number of constraints.
        lower_is_better: Whether the objectives are lower-is-better. Applies to
            all objectives and not to constraints. For constraints, higher is
            better (feasible). Note: Ax allows different metrics to have
            different values of ``lower_is_better``; that isn't enabled for this
            method, but could be.
        observe_noise_sd: Whether the standard deviation of the observation
        noise is observed. Applies to all objective and constraints.
    """
    n_objectives = len(outcome_names) - num_constraints
    metric_cls = BenchmarkMapMetric if use_map_metric else BenchmarkMetric
    objective_metrics = [
        metric_cls(
            name=outcome_names[i],
            lower_is_better=lower_is_better,
            observe_noise_sd=observe_noise_sd,
        )
        for i in range(n_objectives)
    ]
    constraints = _get_constraints(
        constraint_names=outcome_names[n_objectives:],
        observe_noise_sd=observe_noise_sd,
    )
    optimization_config = MultiObjectiveOptimizationConfig(
        objective=MultiObjective(
            objectives=[
                Objective(metric=metric, minimize=lower_is_better)
                for metric in objective_metrics
            ]
        ),
        objective_thresholds=[
            ObjectiveThreshold(
                metric=metric,
                bound=ref_p,
                relative=False,
                op=ComparisonOp.LEQ if metric.lower_is_better else ComparisonOp.GEQ,
            )
            for ref_p, metric in zip(ref_point, objective_metrics, strict=True)
        ],
        outcome_constraints=constraints,
    )
    return optimization_config


def get_continuous_search_space(
    bounds: list[tuple[float, float]], n_dummy_dimensions: int = 0
) -> SearchSpace:
    """
    Create a continuous ``SearchSpace``.

    Args:
        bounds: A list of tuples of lower and upper bounds for each dimension.
            These apply only to the original problem dimensions. Any extra
            dimensions will have bounds [0, 1].
        n_dummy_dimensions: Number of extra dimensions to add to the search
            space.
    """
    original_problem_parameters = [
        RangeParameter(
            name=f"x{i}",
            parameter_type=ParameterType.FLOAT,
            lower=lower,
            upper=upper,
        )
        for i, (lower, upper) in enumerate(bounds)
    ]
    dummy_parameters = [
        RangeParameter(
            name=f"embedding_dummy_{i}",
            parameter_type=ParameterType.FLOAT,
            lower=0,
            upper=1,
        )
        for i in range(n_dummy_dimensions)
    ]
    return SearchSpace(parameters=original_problem_parameters + dummy_parameters)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import warnings
from copy import deepcopy
from typing import Any
from unittest import mock
from unittest.mock import Mock

import numpy as np
import torch
from ax.adapter.base import (
    _combine_multiple_status_quo_observations,
    Adapter,
    clamp_observation_features,
    DataLoaderConfig,
    gen_arms,
    GenResults,
    logger,
    unwrap_observation_data,
)
from ax.adapter.data_utils import ExperimentData, extract_experiment_data
from ax.adapter.factory import get_sobol
from ax.adapter.registry import MBM_X_trans, MBM_X_trans_base, Y_trans
from ax.adapter.transforms.cast import Cast
from ax.adapter.transforms.fill_missing_parameters import FillMissingParameters
from ax.adapter.transforms.standardize_y import StandardizeY
from ax.adapter.transforms.unit_x import UnitX
from ax.core.arm import Arm
from ax.core.base_trial import TrialStatus
from ax.core.experiment import Experiment
from ax.core.map_data import MapData
from ax.core.metric import Metric
from ax.core.objective import Objective, ScalarizedObjective
from ax.core.observation import ObservationData, ObservationFeatures
from ax.core.optimization_config import OptimizationConfig
from ax.core.outcome_constraint import ComparisonOp, OutcomeConstraint
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.parameter_constraint import SumConstraint
from ax.core.search_space import SearchSpace
from ax.core.types import TParameterization
from ax.core.utils import get_target_trial_index
from ax.exceptions.core import UnsupportedError, UserInputError
from ax.exceptions.model import ModelError
from ax.generators.base import Generator
from ax.metrics.branin import BraninMetric
from ax.utils.common.constants import Keys
from ax.utils.common.logger import get_logger
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_data_batch,
    get_branin_experiment,
    get_branin_experiment_with_multi_objective,
    get_branin_experiment_with_timestamp_map_metric,
    get_branin_optimization_config,
    get_experiment,
    get_experiment_with_observations,
    get_map_metric,
    get_non_monolithic_branin_moo_data,
    get_optimization_config_no_constraints,
    get_search_space_for_range_values,
    get_search_space_for_value,
)
from ax.utils.testing.modeling_stubs import (
    get_experiment_for_value,
    get_observation1,
    get_observation1trans,
    get_observation2,
)
from botorch.exceptions.warnings import InputDataWarning
from botorch.models.utils.assorted import validate_input_scaling
from pandas.testing import assert_frame_equal
from pyre_extensions import none_throws

ADAPTER__GEN_PATH: str = "ax.adapter.base.Adapter._gen"


class BaseAdapterTest(TestCase):
    def test_init_empty(self) -> None:
        # Test Adapter initialization with an experiment with no data.
        exp = get_branin_experiment()
        generator = Generator()
        with mock.patch("ax.adapter.base.Adapter._fit") as mock_fit:
            adapter = Adapter(
                experiment=exp, generator=generator, transforms=MBM_X_trans_base
            )
        # Check that the properties are set correctly.
        self.assertEqual(adapter._data_loader_config, DataLoaderConfig())
        self.assertEqual(adapter._raw_transforms, [Cast] + MBM_X_trans_base)
        self.assertEqual(adapter._transform_configs, {})
        self.assertEqual(
            list(adapter.transforms), [t.__name__ for t in [Cast] + MBM_X_trans_base]
        )
        self.assertEqual(adapter.fit_time, adapter.fit_time_since_gen)
        self.assertEqual(adapter._metric_names, set())
        self.assertEqual(adapter._optimization_config, exp.optimization_config)
        self.assertEqual(adapter._training_in_design_idx, [])
        self.assertIsNone(adapter._status_quo)
        self.assertIsNone(adapter._status_quo_name)
        self.assertIsNone(adapter._model_key)
        self.assertIsNone(adapter._model_kwargs)
        self.assertIsNone(adapter._bridge_kwargs)
        self.assertEqual(adapter._search_space, exp.search_space)
        self.assertEqual(adapter._model_space, exp.search_space)
        self.assertTrue(adapter._fit_tracking_metrics)
        self.assertEqual(adapter.outcomes, [])
        self.assertEqual(
            adapter._experiment_has_immutable_search_space_and_opt_config,
            exp.immutable_search_space_and_opt_config,
        )
        self.assertIs(adapter._experiment, exp)
        self.assertEqual(adapter._experiment_properties, exp._properties)
        self.assertEqual(adapter._arms_by_signature, {})
        self.assertIs(adapter.generator, generator)
        # Check that the experiment data object is empty.
        self.assertTrue(adapter._training_data.arm_data.empty)
        self.assertTrue(adapter._training_data.observation_data.empty)
        # Check that fit was called with the transformed arguments.
        search_space = adapter._search_space.clone()
        experiment_data = adapter.get_training_data()
        for t in adapter.transforms.values():
            search_space = t.transform_search_space(search_space)
            experiment_data = t.transform_experiment_data(experiment_data)
        mock_fit.assert_called_with(
            search_space=search_space, experiment_data=experiment_data
        )
        # Test that fit is not called when fit_on_init = False.
        mock_fit.reset_mock()
        adapter = Adapter(experiment=exp, generator=Generator(), fit_on_init=False)
        self.assertEqual(mock_fit.call_count, 0)

    def _test_init_with_data(self, multi_objective: bool) -> None:
        # Test Adapter initialization with a simple experiment with (non-map) data.
        exp_constructor = (
            get_branin_experiment_with_multi_objective
            if multi_objective
            else get_branin_experiment
        )
        exp = exp_constructor(with_completed_batch=True, with_status_quo=True)
        generator = Generator()
        with mock.patch("ax.adapter.base.Adapter._fit") as mock_fit:
            adapter = Adapter(
                experiment=exp, generator=generator, transforms=MBM_X_trans + Y_trans
            )
        # Check that the properties are set correctly.
        # Only checking a subset that are expected to be different than test_init_empty.
        self.assertEqual(adapter._raw_transforms, [Cast] + MBM_X_trans + Y_trans)
        metric_names = set(exp.metrics)
        self.assertEqual(adapter._metric_names, metric_names)
        self.assertEqual(
            adapter._training_in_design_idx, [True] * len(exp.arms_by_name)
        )
        self.assertEqual(adapter._optimization_config, exp.optimization_config)
        self.assertEqual(adapter._status_quo_name, none_throws(exp.status_quo).name)
        # Not checking SQ observation in detail, _set_status_quo is tested separately.
        self.assertIsNotNone(adapter._status_quo)
        self.assertEqual(adapter._arms_by_signature, exp.arms_by_signature)
        # Check the raw training data.
        exp_df = exp.to_df()
        self.assertTrue(
            np.allclose(
                adapter._training_data.arm_data[["x1", "x2"]], exp_df[["x1", "x2"]]
            )
        )
        self.assertTrue(
            np.allclose(
                adapter._training_data.observation_data[
                    [("mean", m) for m in metric_names]
                ],
                exp_df.sort_values(by="arm_name")[list(metric_names)],
            )
        )
        # Check that fit was called with the transformed arguments.
        search_space = adapter._search_space.clone()
        experiment_data = adapter.get_training_data()
        for t in adapter.transforms.values():
            search_space = t.transform_search_space(search_space)
            experiment_data = t.transform_experiment_data(experiment_data)
        mock_fit.assert_called_with(
            search_space=search_space, experiment_data=experiment_data
        )

    def test_init_with_data_single_objective(self) -> None:
        self._test_init_with_data(multi_objective=False)

    def test_init_with_data_multi_objective(self) -> None:
        self._test_init_with_data(multi_objective=True)

    def test_fit_tracking_metrics(self) -> None:
        # Test error when fit_tracking_metrics is False and optimization
        # config is not specified.
        exp = get_branin_experiment(has_optimization_config=False)
        with self.assertRaisesRegex(UserInputError, "fit_tracking_metrics"):
            Adapter(experiment=exp, generator=Generator(), fit_tracking_metrics=False)

        # Test error when fit_tracking_metrics is False and optimization
        # config is updated to include new metrics.
        adapter = Adapter(
            experiment=get_branin_experiment(),
            generator=Generator(),
            fit_tracking_metrics=False,
        )
        new_oc = OptimizationConfig(
            objective=Objective(metric=Metric(name="test_metric2"), minimize=False),
        )
        with self.assertRaisesRegex(UnsupportedError, "fit_tracking_metrics"):
            adapter.gen(n=1, optimization_config=new_oc)

    @mock.patch(
        "ax.adapter.base.gen_arms",
        autospec=True,
        return_value=([Arm(parameters={"x1": 0.5, "x2": 0.5})], None),
    )
    @mock.patch("ax.adapter.base.Adapter._fit", autospec=True)
    def test_gen_base(self, mock_fit: Mock, mock_gen_arms: Mock) -> None:
        transforms = [UnitX, StandardizeY]
        exp = get_branin_experiment(with_completed_trial=True)
        search_space = exp.search_space
        adapter = Adapter(experiment=exp, generator=Generator(), transforms=transforms)

        # Test transforms applied on gen
        mock_return_value = GenResults(
            observation_features=[
                ObservationFeatures(parameters={"x1": 0.0, "x2": 0.0})
            ],
            weights=[1.0],
        )
        oc = get_optimization_config_no_constraints()
        adapter._set_kwargs_to_save(
            model_key="TestModel", model_kwargs={}, bridge_kwargs={}
        )
        # Test input error when generating 0 candidates.
        with self.assertRaisesRegex(UserInputError, "Attempted to generate"):
            adapter.gen(n=0)
        with mock.patch(ADAPTER__GEN_PATH, return_value=mock_return_value) as mock_gen:
            gr = adapter.gen(
                n=1,
                search_space=search_space,
                optimization_config=oc,
                pending_observations={
                    "branin": [ObservationFeatures(parameters={"x1": 10.0, "x2": 15.0})]
                },
                fixed_features=ObservationFeatures({"x1": -5.0}),
            )
        self.assertEqual(gr._model_key, "TestModel")
        tf_search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    name=name,
                    parameter_type=ParameterType.FLOAT,
                    lower=0.0,
                    upper=1.0,
                )
                for name in ("x1", "x2")
            ]
        )
        mock_gen.assert_called_with(
            n=1,
            search_space=tf_search_space,
            optimization_config=oc,
            pending_observations={
                "branin": [ObservationFeatures(parameters={"x1": 1.0, "x2": 1.0})]
            },
            fixed_features=ObservationFeatures({"x1": 0.0}),
            model_gen_options=None,
        )
        mock_gen_arms.assert_called_with(
            arms_by_signature=mock.ANY,
            observation_features=[
                ObservationFeatures(parameters={"x1": -5.0, "x2": 0.0})
            ],
        )

        # Gen with no pending observations and no fixed features
        adapter._optimization_config = None
        with mock.patch(ADAPTER__GEN_PATH, return_value=mock_return_value) as mock_gen:
            adapter.gen(n=1, search_space=search_space, optimization_config=None)
        mock_gen.assert_called_with(
            n=1,
            search_space=tf_search_space,
            optimization_config=None,
            pending_observations={},
            fixed_features=ObservationFeatures(parameters={}),
            model_gen_options=None,
        )

        # Gen with multi-objective optimization config.
        oc2 = OptimizationConfig(
            objective=ScalarizedObjective(
                metrics=[Metric(name="test_metric"), Metric(name="test_metric_2")]
            )
        )
        with mock.patch(ADAPTER__GEN_PATH, return_value=mock_return_value) as mock_gen:
            adapter.gen(n=1, search_space=search_space, optimization_config=oc2)
        mock_gen.assert_called_with(
            n=1,
            search_space=tf_search_space,
            optimization_config=oc2,
            pending_observations={},
            fixed_features=ObservationFeatures(parameters={}),
            model_gen_options=None,
        )

    @mock.patch(
        "ax.adapter.base.Adapter._gen",
        autospec=True,
        return_value=GenResults(
            observation_features=[get_observation1trans().features], weights=[2]
        ),
    )
    def test_gen_on_experiment_with_imm_ss_and_opt_conf(self, _) -> None:
        exp = get_experiment_for_value()
        exp._properties[Keys.IMMUTABLE_SEARCH_SPACE_AND_OPT_CONF] = True
        exp.optimization_config = get_optimization_config_no_constraints()
        adapter = Adapter(experiment=exp, generator=Generator())
        self.assertTrue(adapter._experiment_has_immutable_search_space_and_opt_config)
        gr = adapter.gen(1)
        self.assertIsNone(gr.optimization_config)
        self.assertIsNone(gr.search_space)

    def test_cross_validate_base(self) -> None:
        exp = get_branin_experiment(with_completed_batch=True)
        adapter = Adapter(experiment=exp, generator=Generator(), transforms=[UnitX])
        # Test transforms applied on cross_validate and the warning is suppressed.
        called = False
        mock_predictions: list[ObservationData] = [
            ObservationData(
                metric_names=["branin"],
                means=np.zeros(1),
                covariance=np.ones((1, 1)),
            )
        ]

        def warn_and_return_mock_obs(
            *args: Any, **kwargs: Any
        ) -> list[ObservationData]:
            nonlocal called
            called = True
            validate_input_scaling(
                # Dummy non-scaled values to trigger warnings if checks are enabled.
                train_X=torch.randn(2, 5),
                train_Y=torch.rand(2, 1),
            )
            return mock_predictions

        mock_cv = mock.MagicMock(
            "ax.adapter.base.Adapter._cross_validate",
            autospec=True,
            side_effect=warn_and_return_mock_obs,
        )
        adapter._cross_validate = mock_cv
        cv_training_data = adapter.get_training_data()
        cv_test_points = [ObservationFeatures(parameters={"x1": -5.0, "x2": 0.0})]

        # Test transforms applied on cv_training_data, cv_test_points
        (
            transformed_cv_training_data,
            transformed_cv_test_points,
            transformed_ss,
        ) = adapter._transform_inputs_for_cv(
            cv_training_data=cv_training_data, cv_test_points=cv_test_points
        )
        self.assertEqual(
            transformed_cv_training_data,
            adapter.transforms["UnitX"].transform_experiment_data(
                experiment_data=adapter.transforms["Cast"].transform_experiment_data(
                    experiment_data=deepcopy(cv_training_data)
                )
            ),
        )
        self.assertEqual(
            transformed_cv_test_points,
            [ObservationFeatures(parameters={"x1": 0.0, "x2": 0.0})],
        )
        self.assertEqual(
            transformed_ss,
            SearchSpace(
                parameters=[
                    RangeParameter(
                        name=name,
                        parameter_type=ParameterType.FLOAT,
                        lower=0.0,
                        upper=1.0,
                    )
                    for name in ("x1", "x2")
                ]
            ),
        )

        with warnings.catch_warnings(record=True) as ws:
            cv_predictions = adapter.cross_validate(
                cv_training_data=cv_training_data, cv_test_points=cv_test_points
            )
        self.assertTrue(called)
        self.assertFalse(any(w.category is InputDataWarning for w in ws))

        mock_cv.assert_called_with(
            search_space=transformed_ss,
            cv_training_data=transformed_cv_training_data,
            cv_test_points=cv_test_points,  # in-place untransformed after the call.
            use_posterior_predictive=False,
        )
        self.assertEqual(cv_predictions, mock_predictions)

        # Test use_posterior_predictive in CV
        adapter.cross_validate(
            cv_training_data=cv_training_data,
            cv_test_points=cv_test_points,
            use_posterior_predictive=True,
        )

        mock_cv.assert_called_with(
            search_space=transformed_ss,
            cv_training_data=transformed_cv_training_data,
            cv_test_points=cv_test_points,  # in-place untransformed after the call.
            use_posterior_predictive=True,
        )

    @mock.patch(
        "ax.adapter.base.gen_arms",
        autospec=True,
        return_value=([Arm(parameters={})], None),
    )
    @mock.patch("ax.adapter.base.Adapter._fit", autospec=True)
    def test_repeat_candidates(self, _: Mock, __: Mock) -> None:
        adapter = Adapter(
            experiment=get_experiment_for_value(),
            generator=Generator(),
        )
        # mock _gen to return 1 result
        adapter._gen = mock.MagicMock(
            "ax.adapter.base.Adapter._gen",
            autospec=True,
            return_value=GenResults(
                observation_features=[get_observation1trans().features], weights=[2]
            ),
        )
        adapter._set_kwargs_to_save(
            model_key="TestModel", model_kwargs={}, bridge_kwargs={}
        )
        with self.assertLogs("ax", level="INFO") as cm:
            adapter.gen(
                n=2,
            )
            self.assertTrue(
                any(
                    "was not able to generate 2 unique candidates" in x
                    for x in cm.output
                ),
                cm.output,
            )

        with self.assertLogs("ax", level="INFO") as cm:
            adapter.gen(
                n=1,
            )
            get_logger("ax").info("log to prevent error if there are no other logs")
            self.assertFalse(
                any(
                    "was not able to generate 2 unique candidates" in x
                    for x in cm.output
                ),
                cm.output,
            )

    @mock.patch("ax.adapter.base.Adapter._fit", autospec=True)
    @mock.patch("ax.adapter.base.Adapter._gen", autospec=True)
    def test_timing(self, _: Mock, __: Mock) -> None:
        search_space = get_search_space_for_value()
        experiment = Experiment(search_space=search_space)
        adapter = Adapter(
            experiment=experiment, generator=Generator(), fit_on_init=False
        )
        self.assertEqual(adapter.fit_time, 0.0)
        adapter._fit_if_implemented(
            search_space=search_space,
            experiment_data=adapter._training_data,
            time_so_far=3.0,
        )
        adapter._fit_if_implemented(
            search_space=search_space,
            experiment_data=adapter._training_data,
            time_so_far=2.0,
        )
        adapter._fit_if_implemented(
            search_space=search_space,
            experiment_data=adapter._training_data,
            time_so_far=1.0,
        )
        self.assertAlmostEqual(adapter.fit_time, 6.0, places=1)
        self.assertAlmostEqual(adapter.fit_time_since_gen, 6.0, places=1)
        adapter.gen(1)
        self.assertAlmostEqual(adapter.fit_time, 6.0, places=1)
        self.assertAlmostEqual(adapter.fit_time_since_gen, 0.0, places=1)

    def test_ood_gen(self) -> None:
        # Test fit_out_of_design by returning OOD candidates
        ss = SearchSpace([RangeParameter("x", ParameterType.FLOAT, 0.0, 1.0)])
        experiment = Experiment(search_space=ss)
        adapter = Adapter(
            experiment=experiment,
            generator=Generator(),
            fit_out_of_design=True,
        )
        obs = ObservationFeatures(parameters={"x": 3.0})
        adapter._gen = mock.MagicMock(
            "ax.adapter.base.Adapter._gen",
            autospec=True,
            return_value=GenResults(observation_features=[obs], weights=[2]),
        )
        gr = adapter.gen(n=1)
        self.assertEqual(gr.arms[0].parameters, obs.parameters)

        # Test clamping arms by setting fit_out_of_design=False
        adapter = Adapter(
            experiment=experiment,
            generator=Generator(),
            fit_out_of_design=False,
        )
        obs = ObservationFeatures(parameters={"x": 3.0})
        adapter._gen = mock.MagicMock(
            "ax.adapter.base.Adapter._gen",
            autospec=True,
            return_value=GenResults(observation_features=[obs], weights=[2]),
        )
        gr = adapter.gen(n=1)
        self.assertEqual(gr.arms[0].parameters, {"x": 1.0})

    def test_set_status_quo(self) -> None:
        exp = get_experiment_for_value()
        exp._status_quo = Arm(parameters={"x": 3.0}, name="0_0")
        with self.assertLogs(logger="ax", level="WARNING") as logs:
            adapter = Adapter(experiment=exp, generator=Generator())
        self.assertTrue(
            any("is not present in the training data" in log for log in logs.output)
        )
        # Status quo name is set but status quo itself is not, since there is no data.
        self.assertEqual(adapter.status_quo_name, "0_0")
        self.assertIsNone(adapter.status_quo)

        for num_batch_trial in (1, 2):
            # Experiment with status quo in num_batch_trial trials. Only one completed.
            exp = get_branin_experiment(
                with_batch=True,
                with_status_quo=True,
                num_batch_trial=num_batch_trial,
                with_completed_batch=True,
            )
            adapter = Adapter(experiment=exp, generator=Generator())
            # Status quo is set with the target trial index.
            self.assertEqual(
                none_throws(adapter.status_quo).features.trial_index,
                get_target_trial_index(experiment=exp),
            )
            # Status quo data by trial extracts the data from all trials.
            self.assertEqual(
                set(none_throws(adapter.status_quo_data_by_trial).keys()),
                set(range(num_batch_trial)),
            )
            # Status quo name is set.
            self.assertEqual(adapter._status_quo_name, none_throws(exp.status_quo).name)

    def test_status_quo_for_non_monolithic_data(self) -> None:
        exp = get_branin_experiment_with_multi_objective(with_status_quo=True)
        sobol_generator = get_sobol(
            search_space=exp.search_space,
        )
        sobol_run = sobol_generator.gen(n=5)
        exp.new_batch_trial(
            sobol_run, should_add_status_quo_arm=False
        ).add_status_quo_arm(weight=1.0).run()

        # create data where metrics vary in start and end times
        data = get_non_monolithic_branin_moo_data()
        with mock.patch.object(exp, "lookup_data", return_value=data):
            adapter = Adapter(experiment=exp, generator=Generator(), data=data)
        # Check that SQ is set.
        self.assertEqual(adapter.status_quo_name, "status_quo")
        self.assertIsNotNone(adapter.status_quo)

    def test_set_status_quo_with_multiple_observations(self) -> None:
        # Test for the case where the status quo arm has multiple observations
        # for the target trial. This happens with MapData.

        # Prevent data from being filtered down.
        data_loader_config = DataLoaderConfig(
            latest_rows_per_group=None, fit_only_completed_map_metrics=False
        )

        # Case 1: Experiment has an optimization config with single map key.
        exp = get_branin_experiment_with_timestamp_map_metric(with_status_quo=True)
        # Add a second map metric, and a non-map metric.
        exp.optimization_config = none_throws(exp.optimization_config).clone_with_args(
            outcome_constraints=[
                OutcomeConstraint(
                    metric=get_map_metric("branin_map_constraint"),
                    op=ComparisonOp.LEQ,
                    bound=5.0,
                ),
                OutcomeConstraint(
                    metric=BraninMetric(
                        name="branin_constraint",
                        param_names=["x1", "x2"],
                        lower_is_better=True,
                    ),
                    op=ComparisonOp.LEQ,
                    bound=5.0,
                ),
            ]
        )
        # Attach a trial with status quo & fetch some data.
        exp.new_trial().add_arm(exp.status_quo).run()
        for _ in range(3):
            exp.fetch_data()
        for additional_fetch in (False, True):
            if additional_fetch:
                # Fetch constraint metric an additional time. This will lead to two
                # separate observations for the status quo arm.
                exp.fetch_data(
                    metrics=[exp.metrics["branin_map_constraint"]],
                    combine_with_last_data=True,
                )
            with self.assertNoLogs(logger=logger, level="WARN"), mock.patch(
                "ax.adapter.base._combine_multiple_status_quo_observations",
                wraps=_combine_multiple_status_quo_observations,
            ) as mock_combine:
                adapter = Adapter(
                    experiment=exp,
                    generator=Generator(),
                    data_loader_config=data_loader_config,
                )
            mock_combine.assert_called_once()
            call_kwargs = mock_combine.call_args.kwargs
            # 3 for metric 'branin_map' with timestamp=0, 1, 2, and 1 for metric
            # 'branin' with timestamp=NaN
            self.assertEqual(
                len(call_kwargs["status_quo_observations"]), 4 + additional_fetch
            )
            if additional_fetch:
                # Last observation should only include the constraint metric.
                self.assertEqual(
                    set(call_kwargs["status_quo_observations"][-1].data.metric_names),
                    {"branin_map_constraint"},
                )
            opt_config_metrics = set(none_throws(exp.optimization_config).metrics)
            self.assertEqual(call_kwargs["metrics"], opt_config_metrics)
            adapter_sq = none_throws(adapter.status_quo)
            self.assertEqual(
                adapter_sq.features.parameters, none_throws(exp.status_quo).parameters
            )
            self.assertEqual(
                adapter_sq.features.trial_index, get_target_trial_index(experiment=exp)
            )
            self.assertTrue(
                set(adapter_sq.data.metric_names).issuperset(opt_config_metrics)
            )

        # Case 2: Experiment has an optimization config with no map metrics
        with mock.patch(
            "ax.adapter.base.has_map_metrics", return_value=False
        ) as mock_extract, self.assertLogs(logger=logger, level="WARN") as mock_logs:
            adapter = Adapter(
                experiment=exp,
                generator=Generator(),
                data_loader_config=data_loader_config,
            )
        mock_extract.assert_called_once()
        self.assertIsNone(adapter.status_quo)
        self.assertTrue(
            any(
                "optimization config does not include any MapMetrics" in log
                for log in mock_logs.output
            )
        )

        # Case 3: Experiment doesn't have an optimization config.
        metrics = none_throws(exp.optimization_config).metrics.values()
        exp._optimization_config = None
        # Attach as tracking metric to prevent data filtering.
        for m in metrics:
            exp.add_tracking_metric(m)
        with self.assertLogs(logger=logger, level="WARN") as mock_logs:
            adapter = Adapter(
                experiment=exp,
                generator=Generator(),
                data_loader_config=data_loader_config,
            )
        self.assertIsNone(adapter.status_quo)
        self.assertTrue(
            any(
                "the Adapter does not have an optimization config" in log
                for log in mock_logs.output
            )
        )

    def test_transform_observations(self) -> None:
        """
        This functionality is unused, even in the subclass where it is implemented.
        """
        adapter = Adapter(experiment=get_experiment_for_value(), generator=Generator())
        with self.assertRaises(NotImplementedError):
            adapter.transform_observations([])
        with self.assertRaises(NotImplementedError):
            adapter.transform_observations([])

    def test_UnwrapObservationData(self) -> None:
        observation_data = [get_observation1().data, get_observation2().data]
        f, cov = unwrap_observation_data(observation_data)
        self.assertEqual(f["a"], [2.0, 2.0])
        self.assertEqual(f["b"], [4.0, 1.0])
        self.assertEqual(cov["a"]["a"], [1.0, 2.0])
        self.assertEqual(cov["b"]["b"], [4.0, 5.0])
        self.assertEqual(cov["a"]["b"], [2.0, 3.0])
        self.assertEqual(cov["b"]["a"], [3.0, 4.0])
        # Check that errors if metric mismatch
        od3 = ObservationData(
            metric_names=["a"], means=np.array([2.0]), covariance=np.array([[4.0]])
        )
        with self.assertRaises(ValueError):
            unwrap_observation_data(observation_data + [od3])

    def test_gen_arms(self) -> None:
        p1: TParameterization = {"x": 0, "y": 1}
        p2: TParameterization = {"x": 4, "y": 8}
        observation_features = [
            ObservationFeatures(parameters=p1),
            ObservationFeatures(parameters=p2),
        ]
        arms, candidate_metadata = gen_arms(observation_features=observation_features)
        self.assertEqual(arms[0].parameters, p1)
        self.assertIsNone(candidate_metadata)

        arm = Arm(name="1_1", parameters=p1)
        arms_by_signature = {arm.signature: arm}
        observation_features[0].metadata = {"some_key": "some_val_0"}
        observation_features[1].metadata = {"some_key": "some_val_1"}
        arms, candidate_metadata = gen_arms(
            observation_features=observation_features,
            arms_by_signature=arms_by_signature,
        )
        self.assertEqual(arms[0].name, "1_1")
        self.assertEqual(
            candidate_metadata,
            {
                arms[0].signature: {"some_key": "some_val_0"},
                arms[1].signature: {"some_key": "some_val_1"},
            },
        )

    def test_ClampObservationFeaturesNearBounds(self) -> None:
        cases = [
            (
                ObservationFeatures(
                    parameters={"w": 1.0, "x": 2, "y": "foo", "z": True}
                ),
                ObservationFeatures(
                    parameters={"w": 1.0, "x": 2, "y": "foo", "z": True}
                ),
            ),
            (
                ObservationFeatures(
                    parameters={"w": 0.0, "x": 2, "y": "foo", "z": True}
                ),
                ObservationFeatures(
                    parameters={"w": 0.5, "x": 2, "y": "foo", "z": True}
                ),
            ),
            (
                ObservationFeatures(
                    parameters={"w": 100.0, "x": 2, "y": "foo", "z": True}
                ),
                ObservationFeatures(
                    parameters={"w": 5.5, "x": 2, "y": "foo", "z": True}
                ),
            ),
            (
                ObservationFeatures(
                    parameters={"w": 1.0, "x": 0, "y": "foo", "z": True}
                ),
                ObservationFeatures(
                    parameters={"w": 1.0, "x": 1, "y": "foo", "z": True}
                ),
            ),
            (
                ObservationFeatures(
                    parameters={"w": 1.0, "x": 11, "y": "foo", "z": True}
                ),
                ObservationFeatures(
                    parameters={"w": 1.0, "x": 10, "y": "foo", "z": True}
                ),
            ),
        ]
        search_space = get_experiment().search_space
        for obs_ft, expected_obs_ft in cases:
            actual_obs_ft = clamp_observation_features([obs_ft], search_space)
            self.assertEqual(actual_obs_ft[0], expected_obs_ft)

    @mock.patch("ax.adapter.base.Adapter._fit", autospec=True)
    def test_FillMissingParameters(self, mock_fit: Mock) -> None:
        # Create experiment with arms from two search spaces
        ss1 = SearchSpace(
            parameters=[
                RangeParameter(
                    name="x1", parameter_type=ParameterType.FLOAT, lower=-5, upper=10
                )
            ],
        )
        ss2 = SearchSpace(
            parameters=[
                RangeParameter(
                    name="x1", parameter_type=ParameterType.FLOAT, lower=-5, upper=10
                ),
                RangeParameter(
                    name="x2", parameter_type=ParameterType.FLOAT, lower=0, upper=15
                ),
            ],
        )
        sq_arm = Arm(name="status_quo", parameters={"x1": None, "x2": None})
        experiment = Experiment(
            name="test",
            search_space=ss2,
            optimization_config=get_branin_optimization_config(),
            status_quo=sq_arm,
            is_test=True,
        )
        generator1 = get_sobol(search_space=ss1)
        gr1 = generator1.gen(n=5)
        generator2 = get_sobol(search_space=ss2)
        gr2 = generator2.gen(n=5)
        sq_vals = {"x1": 5.0, "x2": 5.0}
        for gr in [gr1, gr2]:
            trial = experiment.new_batch_trial(should_add_status_quo_arm=True)
            trial.add_generator_run(gr)
            trial.mark_running(no_runner_required=True)
            trial.mark_completed()
            experiment.attach_data(
                get_branin_data_batch(batch=trial, fill_vals=sq_vals)
            )
        # Fit model without filling missing parameters
        m = Adapter(experiment=experiment, generator=Generator())
        self.assertEqual(
            [t.__name__ for t in m._raw_transforms],  # pyre-ignore[16]
            ["Cast"],
        )
        # Check that SQ and all trial 1 are OOD
        ood_arms = set(
            m.get_training_data()
            .arm_data.loc[~np.array(m.training_in_design)]
            .index.get_level_values("arm_name")
        )
        self.assertEqual(
            set(ood_arms), {"status_quo", "0_0", "0_1", "0_2", "0_3", "0_4"}
        )
        # Fit with filling missing parameters
        m = Adapter(
            experiment=experiment,
            generator=Generator(),
            search_space=ss2,
            transforms=[],  # FillMissingParameters added by default.
            transform_configs={"FillMissingParameters": {"fill_values": sq_vals}},
        )
        self.assertEqual(
            [t.__name__ for t in m._raw_transforms], ["FillMissingParameters", "Cast"]
        )
        # All arms are in design now
        self.assertEqual(sum(m.training_in_design), 12)
        # Check the arms with missing values were correctly filled
        fit_args = mock_fit.call_args.kwargs
        arm_data = fit_args["experiment_data"].arm_data
        sq_params = arm_data.loc[
            arm_data.index.get_level_values("arm_name") == "status_quo"
        ][["x1", "x2"]].to_dict(orient="records")
        self.assertEqual(sq_params, [sq_vals, sq_vals])
        trial_0_params = arm_data.loc[(0, slice(None))]
        self.assertEqual(
            trial_0_params["x2"].to_list(), [sq_vals["x2"]] * len(trial_0_params)
        )

    def test_set_model_space(self) -> None:
        # Set up experiment
        experiment = get_branin_experiment()
        # SQ values are OOD
        sq_vals = {"x1": 5.0, "x2": 20.0}
        # SQ is specified OOD
        experiment.status_quo = Arm(
            name="status_quo", parameters={"x1": None, "x2": None}
        )
        gr = get_sobol(search_space=experiment.search_space).gen(n=5)
        trial = experiment.new_batch_trial()
        trial.add_generator_run(gr)
        trial.add_arm(Arm(name="custom", parameters={"x1": -20, "x2": 18.0}))
        trial.add_arm(experiment.status_quo)
        trial.mark_running(no_runner_required=True)
        experiment.attach_data(get_branin_data_batch(batch=trial, fill_vals=sq_vals))
        trial.mark_completed()
        # Make search space with a parameter constraint
        ss = experiment.search_space.clone()
        ss.set_parameter_constraints(
            [
                SumConstraint(
                    parameters=list(ss.parameters.values()),
                    is_upper_bound=True,
                    bound=30.0,
                )
            ]
        )

        # Check that SQ and custom are OOD
        m = Adapter(
            experiment=experiment,
            generator=Generator(),
            search_space=ss,
            expand_model_space=False,
        )
        ood_arms = set(
            m.get_training_data()
            .arm_data.loc[~np.array(m.training_in_design)]
            .index.get_level_values("arm_name")
        )
        self.assertEqual(set(ood_arms), {"status_quo", "custom"})
        self.assertEqual(m.model_space.parameters["x1"].lower, -5.0)  # pyre-ignore[16]
        self.assertEqual(m.model_space.parameters["x2"].upper, 15.0)  # pyre-ignore[16]
        self.assertEqual(len(m.model_space.parameter_constraints), 1)

        # With expand model space, custom is not OOD, and model space is expanded
        m = Adapter(
            experiment=experiment,
            generator=Generator(),
            search_space=ss,
        )
        ood_arms = set(
            m.get_training_data()
            .arm_data.loc[~np.array(m.training_in_design)]
            .index.get_level_values("arm_name")
        )
        self.assertEqual(set(ood_arms), {"status_quo"})
        self.assertEqual(m.model_space.parameters["x1"].lower, -20.0)
        self.assertEqual(m.model_space.parameters["x2"].upper, 18.0)
        self.assertEqual(m.model_space.parameter_constraints, [])

        # With fill values, SQ is also in design, and x2 is further expanded
        m = Adapter(
            experiment=experiment,
            generator=Generator(),
            search_space=ss,
            transforms=[FillMissingParameters],
            transform_configs={"FillMissingParameters": {"fill_values": sq_vals}},
        )
        self.assertEqual(sum(m.training_in_design), 7)
        self.assertEqual(m.model_space.parameters["x2"].upper, 20)
        self.assertEqual(m.model_space.parameter_constraints, [])

    @mock.patch(
        "ax.adapter.base.extract_experiment_data", wraps=extract_experiment_data
    )
    def test_fit_only_completed_map_metrics(
        self, mock_extract_experiment_data: Mock
    ) -> None:
        # _prepare_observations is called in the constructor and itself calls
        # observations_from_data with expanded statuses to include.
        experiment = get_experiment_for_value()
        experiment.status_quo = Arm(name="1_1", parameters={"x": 3.0})
        data_loader_config = DataLoaderConfig(fit_only_completed_map_metrics=False)
        Adapter(
            experiment=experiment,
            generator=Generator(),
            data=MapData(),
            data_loader_config=data_loader_config,
        )
        kwargs = mock_extract_experiment_data.call_args.kwargs
        self.assertEqual(kwargs["data_loader_config"], data_loader_config)
        mock_extract_experiment_data.reset_mock()

        # With fit_only_completed_map_metrics=True, statuses to fit is limited.
        data_loader_config = DataLoaderConfig(fit_only_completed_map_metrics=True)
        Adapter(
            experiment=experiment,
            generator=Generator(),
            data_loader_config=data_loader_config,
        )
        kwargs = mock_extract_experiment_data.call_args.kwargs
        self.assertEqual(kwargs["data_loader_config"], data_loader_config)
        self.assertEqual(
            data_loader_config.statuses_to_fit_map_metric, {TrialStatus.COMPLETED}
        )

    def test_data_extraction_from_experiment(self) -> None:
        # Checks that data is extracted from experiment both on __init__ and
        # in _process_and_transform_data, if it is not provided.
        exp = get_experiment_for_value()
        lookup_patch = mock.patch.object(
            exp, "lookup_data", return_value=exp.lookup_data()
        ).start()
        adapter = Adapter(experiment=exp, generator=Generator())
        lookup_patch.assert_called_once()
        lookup_patch.reset_mock()
        adapter._process_and_transform_data(experiment=exp)
        lookup_patch.assert_called_once()
        lookup_patch.reset_mock()
        # Not called if data is provided.
        adapter = Adapter(experiment=exp, generator=Generator(), data=MapData())
        adapter._process_and_transform_data(experiment=exp, data=MapData())
        lookup_patch.assert_not_called()

    def test_predict(self) -> None:
        # Construct an experiment with observations having std dev = 2.0
        np_obs = np.random.randn(5, 1)
        np_obs = np_obs / np.std(np_obs, ddof=1) * 2.0
        np_obs = np_obs - np.mean(np_obs)

        search_space = get_search_space_for_range_values(min=2.0, max=6.0)
        experiment = get_experiment_with_observations(
            observations=np_obs.tolist(), search_space=search_space
        )
        adapter = Adapter(
            experiment=experiment,
            generator=Generator(),
            transforms=[UnitX, StandardizeY],
        )
        obs_features = [
            ObservationFeatures(parameters={"x": 3.0, "y": 4.0}) for _ in range(3)
        ]
        tf_obs_features = UnitX(
            search_space=search_space
        ).transform_observation_features(observation_features=obs_features)

        # Test prediction with arms.
        with self.assertRaisesRegex(
            UserInputError, "Input to predict must be a list of `ObservationFeatures`."
        ):
            # pyre-ignore[6]: Intentionally wrong argument type.
            adapter.predict([Arm(parameters={"x": 1.0})])

        # Test errors on prediction.
        adapter._predict = mock.MagicMock(
            "ax.adapter.base.Adapter._predict",
            autospec=True,
            side_effect=ValueError("Predict failed"),
        )
        with self.assertRaisesRegex(ValueError, "Predict failed"):
            adapter.predict(observation_features=obs_features)

        def mock_predict(
            observation_features: list[ObservationFeatures],
            use_posterior_predictive: bool = False,
        ) -> list[ObservationData]:
            return [
                ObservationData(
                    metric_names=["m1"], means=np.ones((1)), covariance=np.ones((1, 1))
                )
                for _ in observation_features
            ]

        with mock.patch.object(
            adapter, "_predict", side_effect=mock_predict
        ) as mock_pred:
            f, _ = adapter.predict(observation_features=obs_features)
        # Check that _predict was called with transformed features.
        mock_pred.assert_called_once_with(
            observation_features=tf_obs_features, use_posterior_predictive=False
        )
        # Check that the predictions were un-transformed with std dev = 2.0.
        self.assertTrue(np.allclose(f["m1"], np.ones(3) * 2.0))

        # Test for error if an observation is dropped.
        with mock.patch.object(
            adapter, "_predict", side_effect=mock_predict
        ), self.assertRaisesRegex(ModelError, "Predictions resulted in fewer"):
            adapter.predict(
                observation_features=[
                    ObservationFeatures(parameters={"x": 3.0, "y": 4.0}),
                    ObservationFeatures(parameters={"x": 3.0, "y": None}),
                ]
            )

    def test_get_training_data(self) -> None:
        # Construct experiment with some out of design training data.
        # Search space is x, y; both are range parameters in [0, 1].
        experiment = get_experiment_with_observations(
            observations=[[1.0], [2.0], [3.0], [4.0]],
            parameterizations=[
                {"x": 0.0, "y": 0.0},
                {"x": 2.0, "y": 1.0},
                {"x": 0.5, "y": 1.0},
                {"x": 0.5},
            ],
        )
        adapter = Adapter(
            experiment=experiment, generator=Generator(), expand_model_space=False
        )
        # Check the default behavior, includes all data.
        training_data = adapter.get_training_data()
        self.assertIsInstance(training_data, ExperimentData)
        self.assertEqual(len(training_data.arm_data), 4)
        # Check in-design only.
        in_design_training_data = adapter.get_training_data(filter_in_design=True)
        assert_frame_equal(
            in_design_training_data.arm_data, training_data.arm_data.iloc[[0, 2]]
        )
        assert_frame_equal(
            in_design_training_data.observation_data,
            training_data.observation_data.iloc[[0, 2]],
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from typing import Callable, TYPE_CHECKING

import numpy as np
import numpy.typing as npt
from ax.adapter.data_utils import ExperimentData
from ax.adapter.transforms.base import Transform
from ax.adapter.transforms.log_y import match_ci_width
from ax.core.observation import Observation, ObservationData
from ax.core.search_space import SearchSpace
from ax.generators.types import TConfig

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401


class BilogY(Transform):
    """Apply a bilog-transform to the outcome constraint metrics.

    The bilog transform was introduced in
    https://proceedings.mlr.press/v130/eriksson21a/eriksson21a.pdf, with the key idea
    being to magnify the region around the outcome constraint constraint boundary,
    thus making it easier to model that part of the metric space. We use a matching
    procedure based on the width of the CIs to transform thevariance observations.
    This transform is applied in-place.

    We define the transform as: f(y) = bound + sign(y - bound) * log(|y - bound| + 1)
    where bound is the outcome constraint bound. This has the properties:
        1. It magnifies the region around y=bound, which makes it easier to
            model the constraint boundary.
        2. f(bound) = bound, so we don't have to modify the optimization config.
    """

    def __init__(
        self,
        search_space: SearchSpace | None = None,
        observations: list[Observation] | None = None,
        experiment_data: ExperimentData | None = None,
        adapter: adapter_module.base.Adapter | None = None,
        config: TConfig | None = None,
    ) -> None:
        """Initialize the ``BilogY`` transform.

        Args:
            search_space: The search space of the experiment.
            observations: A list of observations from the experiment.
            experiment_data: A container for the parameterizations, metadata and
                observations for the trials in the experiment.
                Constructed using ``extract_experiment_data``.
            adapter: Adapter for referencing experiment, status quo, etc.
            config: A dictionary of options specific to each transform.
        """
        super().__init__(
            search_space=search_space,
            observations=observations,
            experiment_data=experiment_data,
            adapter=adapter,
            config=config,
        )
        if adapter is not None and adapter._optimization_config is not None:
            # TODO @deriksson: Add support for relative outcome constraints
            self.metric_to_bound: dict[str, float] = {
                oc.metric.name: oc.bound
                for oc in adapter._optimization_config.outcome_constraints
                if not oc.relative
            }
        else:
            self.metric_to_bound = {}

    def _transform_observation_data(
        self,
        observation_data: list[ObservationData],
    ) -> list[ObservationData]:
        """Apply bilog to observation data in place."""
        return self._reusable_transform(
            observation_data=observation_data,
            transform=bilog_transform,
        )

    def _untransform_observation_data(
        self,
        observation_data: list[ObservationData],
    ) -> list[ObservationData]:
        """Apply inverse bilog to observation data in place."""
        return self._reusable_transform(
            observation_data=observation_data,
            transform=inv_bilog_transform,
        )

    def _reusable_transform(
        self,
        observation_data: list[ObservationData],
        transform: Callable[[npt.NDArray, npt.NDArray], npt.NDArray],
    ) -> list[ObservationData]:
        for obsd in observation_data:
            for i, m in enumerate(obsd.metric_names):
                if m in self.metric_to_bound.keys():
                    bound = self.metric_to_bound[m]
                    obsd.means[i], obsd.covariance[i, i] = match_ci_width(
                        mean=obsd.means[i],
                        sem=None,
                        variance=obsd.covariance[i, i],
                        transform=lambda y, bound=bound: transform(y, bound),
                    )
        return observation_data

    def transform_experiment_data(
        self, experiment_data: ExperimentData
    ) -> ExperimentData:
        obs_data = experiment_data.observation_data
        metric_names = experiment_data.metric_names
        for metric, bound in self.metric_to_bound.items():
            if metric not in metric_names:
                continue
            obs_data[("mean", metric)], obs_data[("sem", metric)] = match_ci_width(
                mean=obs_data[("mean", metric)].to_numpy(),
                sem=obs_data[("sem", metric)].to_numpy(),
                variance=None,
                transform=lambda y, bound=bound: bilog_transform(y, bound),
            )
        return ExperimentData(
            arm_data=experiment_data.arm_data, observation_data=obs_data
        )


def bilog_transform(y: npt.NDArray, bound: npt.NDArray | float) -> npt.NDArray:
    """Bilog transform: f(y) = bound + sign(y - bound) * log(|y - bound| + 1)"""
    diff = y - bound
    return bound + np.sign(diff) * np.log(np.abs(diff) + 1)


def inv_bilog_transform(y: npt.NDArray, bound: npt.NDArray | float) -> npt.NDArray:
    """Inverse bilog transform: f(y) = bound + sign(y - bound) * expm1(|y - bound|)"""
    diff = y - bound
    sign = np.sign(diff)
    return bound + sign * np.expm1(diff * sign)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy
from math import nan
from unittest import mock

import numpy as np

from ax.adapter.data_utils import DataLoaderConfig, extract_experiment_data
from ax.adapter.registry import Generators
from ax.core.data import Data
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.core.trial_status import NON_ABANDONED_STATUSES, TrialStatus
from ax.exceptions.core import UnsupportedError
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_experiment,
    get_branin_experiment_with_multi_objective,
    get_branin_experiment_with_timestamp_map_metric,
    get_experiment_with_observations,
)
from pandas import DataFrame, MultiIndex, Timestamp
from pandas.testing import assert_frame_equal


class TestDataUtils(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.maxDiff = None

    def test_data_loader_config(self) -> None:
        # Defaults
        config = DataLoaderConfig()
        self.assertFalse(config.fit_out_of_design)
        self.assertFalse(config.fit_abandoned)
        self.assertFalse(config.fit_only_completed_map_metrics)
        self.assertEqual(config.latest_rows_per_group, 1)
        self.assertIsNone(config.limit_rows_per_group)
        self.assertIsNone(config.limit_rows_per_metric)
        self.assertEqual(config.statuses_to_fit, NON_ABANDONED_STATUSES)
        self.assertEqual(config.statuses_to_fit_map_metric, NON_ABANDONED_STATUSES)
        # Validation for latest / limit rows.
        with self.assertRaisesRegex(UnsupportedError, "must be None if either of"):
            DataLoaderConfig(latest_rows_per_group=1, limit_rows_per_metric=5)
        # With a bunch of modifications.
        config = DataLoaderConfig(
            fit_out_of_design=True,
            fit_abandoned=True,
            fit_only_completed_map_metrics=False,
            latest_rows_per_group=None,
            limit_rows_per_metric=10,
            limit_rows_per_group=20,
        )
        self.assertTrue(config.fit_out_of_design)
        self.assertTrue(config.fit_abandoned)
        self.assertFalse(config.fit_only_completed_map_metrics)
        self.assertIsNone(config.latest_rows_per_group)
        self.assertEqual(config.limit_rows_per_metric, 10)
        self.assertEqual(config.limit_rows_per_group, 20)
        self.assertEqual(config.statuses_to_fit, set(TrialStatus))
        self.assertEqual(config.statuses_to_fit_map_metric, set(TrialStatus))

    def test_extract_experiment_data_empty(self) -> None:
        # Tests extraction of experiment data from experiments with no data.
        empty_exp = get_branin_experiment()
        for is_map in [False, True]:
            empty_exp = (
                get_branin_experiment_with_timestamp_map_metric()
                if is_map
                else get_branin_experiment()
            )
            experiment_data = extract_experiment_data(
                experiment=empty_exp, data_loader_config=DataLoaderConfig()
            )
            expected_ind_names = ["trial_index", "arm_name"]
            self.assertEqual(len(experiment_data.arm_data), 0)
            self.assertEqual(experiment_data.arm_data.index.names, expected_ind_names)
            if is_map:
                expected_ind_names.append("step")
            self.assertEqual(len(experiment_data.observation_data), 0)
            self.assertEqual(
                experiment_data.observation_data.index.names, expected_ind_names
            )
            self.assertEqual(
                experiment_data.arm_data.columns.to_list(),
                list(empty_exp.parameters) + ["metadata"],
            )
            self.assertEqual(experiment_data, experiment_data)

    def test_extract_experiment_data_non_map(self) -> None:
        # This is a 2 objective experiment with 2 trials, 1 arm each.
        observations = [[0.1, 1.0], [0.2, 2.0]]
        exp = get_experiment_with_observations(observations=observations)
        # Add another trial but fail it.
        # Also add some custom arm metadata.
        sobol = Generators.SOBOL(experiment=exp)
        exp.new_trial(generator_run=sobol.gen(1)).run().mark_failed()
        # Add an abandoned trial but include data for one metric.
        # Also add some custom arm metadata.
        gr = sobol.gen(1)
        gr._candidate_metadata_by_arm_signature = {
            gr.arms[0].signature: {"test": "test_metadata"}
        }
        t = exp.new_trial(generator_run=gr).mark_abandoned()
        data = Data(
            df=DataFrame.from_records(
                [
                    {
                        "arm_name": t.arms[0].name,
                        "metric_name": "m1",
                        "mean": 0.4,
                        "sem": 0.2,
                        "trial_index": t.index,
                    }
                ]
            )
        )
        exp.attach_data(data)

        # Test with default config.
        experiment_data = extract_experiment_data(
            experiment=exp, data_loader_config=DataLoaderConfig()
        )
        # Arm data: Only the trials that have data and are in valid statuses
        # should be included. This excludes arm 3_0 since it is ABANDONED.
        # Filtering happens when constructing datasets.
        valid_arms = ["0_0", "1_0"]
        expected_arm_df = DataFrame(
            [exp.arms_by_name[arm_name].parameters for arm_name in valid_arms],
            index=MultiIndex.from_tuples(
                [(0, "0_0"), (1, "1_0")],
                names=["trial_index", "arm_name"],
            ),
        )
        assert_frame_equal(
            experiment_data.arm_data.drop("metadata", axis=1), expected_arm_df
        )
        # Check metadata. It only includes info about trial completion etc.
        metadata = experiment_data.arm_data["metadata"].tolist()
        self.assertEqual(
            metadata,
            [
                {Keys.TRIAL_COMPLETION_TIMESTAMP: mock.ANY},
                {Keys.TRIAL_COMPLETION_TIMESTAMP: mock.ANY},
            ],
        )
        # Observation data: Only completed trials should be included.
        # First 4 rows correspond to 2 metrics from the 2 completed trials.
        data_df = exp.lookup_data().df[:4]
        expected_obs_df = data_df.pivot(
            columns="metric_name",
            index=["trial_index", "arm_name"],
            values=["mean", "sem"],
        )
        assert_frame_equal(experiment_data.observation_data, expected_obs_df)
        self.assertTrue(
            experiment_data.observation_data.index.equals(
                MultiIndex.from_tuples(
                    [(0, "0_0"), (1, "1_0")],
                    names=["trial_index", "arm_name"],
                )
            )
        )
        self.assertTrue(
            experiment_data.observation_data.columns.equals(
                MultiIndex.from_tuples(
                    [("mean", "m1"), ("mean", "m2"), ("sem", "m1"), ("sem", "m2")],
                    names=[None, "metric_name"],
                )
            )
        )
        # The successful trials don't have SEMs.
        self.assertTrue(experiment_data.observation_data["sem"].isnull().all().all())
        # Check the metric values.
        self.assertEqual(
            experiment_data.observation_data["mean"].to_numpy().tolist(), observations
        )
        # Check equality with self.
        self.assertEqual(experiment_data, experiment_data)

        # Test with config that includes abandoned trials.
        experiment_data = extract_experiment_data(
            experiment=exp, data_loader_config=DataLoaderConfig(fit_abandoned=True)
        )
        # Arm data now includes 3_0 since fit_abandoned=True.
        valid_arms = ["0_0", "1_0", "3_0"]
        expected_arm_df = DataFrame(
            [exp.arms_by_name[arm_name].parameters for arm_name in valid_arms],
            index=MultiIndex.from_tuples(
                [(0, "0_0"), (1, "1_0"), (3, "3_0")],
                names=["trial_index", "arm_name"],
            ),
        )
        assert_frame_equal(
            experiment_data.arm_data.drop("metadata", axis=1), expected_arm_df
        )
        # Check metadata. It only includes info about trial completion etc.
        metadata = experiment_data.arm_data["metadata"].tolist()
        self.assertEqual(
            metadata,
            [
                {Keys.TRIAL_COMPLETION_TIMESTAMP: mock.ANY},
                {Keys.TRIAL_COMPLETION_TIMESTAMP: mock.ANY},
                {Keys.TRIAL_COMPLETION_TIMESTAMP: mock.ANY, "test": "test_metadata"},
            ],
        )
        # All data should be included.
        data_df = exp.lookup_data().df
        expected_obs_df = data_df.pivot(
            columns="metric_name",
            index=["trial_index", "arm_name"],
            values=["mean", "sem"],
        )
        assert_frame_equal(experiment_data.observation_data, expected_obs_df)
        self.assertEqual(len(experiment_data.observation_data), 3)
        # Check the metric & sem values.
        NAN = float("nan")
        self.assertTrue(
            np.array_equal(
                experiment_data.observation_data["mean"].to_numpy(),
                np.asarray(observations + [[0.4, NAN]]),
                equal_nan=True,
            )
        )
        self.assertTrue(
            np.array_equal(
                experiment_data.observation_data["sem"].to_numpy(),
                np.asarray([[NAN, NAN], [NAN, NAN], [0.2, NAN]]),
                equal_nan=True,
            )
        )

    def test_extract_experiment_data_map(self) -> None:
        exp = get_branin_experiment_with_timestamp_map_metric(with_trials_and_data=True)
        t_0_metric = 55.602112642270264
        t_1_metric = 27.702905548512433
        # Test with default config.
        experiment_data = extract_experiment_data(
            experiment=exp,
            data_loader_config=DataLoaderConfig(fit_only_completed_map_metrics=True),
        )
        # Arm data: First two trials should be included, since they have data.
        expected_arm_df = DataFrame(
            [{"x1": 0.0, "x2": 0.0}, {"x1": 1.0, "x2": 1.0}],
            index=MultiIndex.from_tuples(
                [(0, "0_0"), (1, "1_0")], names=["trial_index", "arm_name"]
            ),
        )
        assert_frame_equal(
            experiment_data.arm_data.drop("metadata", axis=1), expected_arm_df
        )
        # Observation data: By default only includes completed map metrics.
        # There is none, so map metrics are not included.
        metrics = set(experiment_data.metric_names)
        self.assertEqual(metrics, {"branin"})
        self.assertEqual(len(experiment_data.observation_data), 2)
        # Complete a trial to include map metrics.
        exp.trials[0].complete()
        experiment_data = extract_experiment_data(
            experiment=exp,
            data_loader_config=DataLoaderConfig(fit_only_completed_map_metrics=True),
        )
        # Arm data is not changed.
        assert_frame_equal(
            experiment_data.arm_data.drop("metadata", axis=1), expected_arm_df
        )
        # Observation data: Map metrics should be included but only with latest
        # timestamp for trial 0.
        metrics = set(experiment_data.metric_names)
        self.assertEqual(metrics, {"branin", "branin_map"})
        index = MultiIndex.from_tuples(
            [(0, "0_0", float("NaN")), (0, "0_0", 3.0), (1, "1_0", float("NaN"))],
            names=["trial_index", "arm_name", "step"],
        )
        expected_mean_df = DataFrame(
            [
                {"branin": t_0_metric, "branin_map": None},
                {"branin": None, "branin_map": t_0_metric},
                {"branin": t_1_metric, "branin_map": None},
            ],
            index=index,
        )
        self.assertTrue(
            experiment_data.observation_data["mean"].equals(expected_mean_df)
        )
        expected_sem_df = DataFrame(
            [
                {"branin": 0.0, "branin_map": None},
                {"branin": None, "branin_map": 0.0},
                {"branin": 0.0, "branin_map": None},
            ],
            index=index,
        )
        self.assertTrue(experiment_data.observation_data["sem"].equals(expected_sem_df))

        # Test with config that includes all map data.
        experiment_data = extract_experiment_data(
            experiment=exp,
            data_loader_config=DataLoaderConfig(
                fit_only_completed_map_metrics=False,
                latest_rows_per_group=None,
            ),
        )
        # Arm data is not changed.
        assert_frame_equal(
            experiment_data.arm_data.drop("metadata", axis=1), expected_arm_df
        )
        # Observation data: Map metrics should be included for all timestamps.
        metrics = set(experiment_data.metric_names)
        self.assertEqual(metrics, {"branin", "branin_map"})
        index = MultiIndex.from_tuples(
            [
                (0, "0_0", nan),
                (0, "0_0", 0.0),
                (0, "0_0", 1.0),
                (0, "0_0", 2.0),
                (0, "0_0", 3.0),
                (1, "1_0", nan),
                (1, "1_0", 0.0),
                (1, "1_0", 1.0),
            ],
            names=["trial_index", "arm_name", "step"],
        )
        expected_mean_df = DataFrame(
            [
                {"branin": t_0_metric, "branin_map": None},  # t=nan
                {"branin": None, "branin_map": t_0_metric},  # t=0
                {"branin": None, "branin_map": t_0_metric},  # t=1
                {"branin": None, "branin_map": t_0_metric},  # t=2
                {"branin": None, "branin_map": t_0_metric},  # t=3
                {"branin": t_1_metric, "branin_map": None},  # t=nan
                {"branin": None, "branin_map": t_1_metric},  # t=0
                {"branin": None, "branin_map": t_1_metric},  # t=1
            ],
            index=index,
        )
        self.assertTrue(
            experiment_data.observation_data["mean"].equals(expected_mean_df)
        )
        # Check equality with self.
        self.assertEqual(experiment_data, experiment_data)

    def test_extract_experiment_data_batch_trials(self) -> None:
        # Check that abandoned arms are correctly handled in BatchTrial.
        experiment = get_branin_experiment(with_batch=True, num_batch_trial=3)
        # Add data for all trials.
        experiment.trials[0].mark_completed(unsafe=True)
        experiment.trials[1].run()
        experiment.trials[2].run()
        experiment.fetch_data()
        # Abandon trial 1 and some arms of trial 2.
        experiment.trials[1].mark_abandoned(unsafe=True)
        experiment.trials[2].mark_arm_abandoned(arm_name="2_14")
        experiment.trials[2].mark_arm_abandoned(arm_name="2_13")
        experiment.trials[2].mark_arm_abandoned(arm_name="2_12")
        # We expect to see only trial 0 and non-abandoned arms of trial 2.
        experiment_data = extract_experiment_data(
            experiment=experiment, data_loader_config=DataLoaderConfig()
        )
        expected_arms = {
            arm.name
            for arm in experiment.trials[0].arms + experiment.trials[2].active_arms
        }
        for df in [experiment_data.arm_data, experiment_data.observation_data]:
            self.assertEqual(set(df.index.get_level_values("arm_name")), expected_arms)

    def test_extract_experiment_data_with_metadata_columns(self) -> None:
        # Tests the case where the Data.df includes additional columns,
        # such as start_time and end_time, besides the usual required columns.
        # In this case, observation_data will include additional columns like
        # (metadata, start_time) and (metadata, end_time).
        exp = get_branin_experiment_with_multi_objective(with_trial=True, num_trial=3)
        # Add data with start_time and end_time only for one of the metrics.
        data = Data(
            df=DataFrame.from_records(
                [
                    {
                        "arm_name": t.arms[0].name,
                        "metric_name": "branin_a",
                        "mean": 0.4 * t.index,
                        "sem": 0.2 + 0.1 * t.index,
                        "trial_index": t.index,
                        "start_time": float(t.index),
                        "end_time": t.index + 5.0,
                    }
                    for t in exp.trials.values()
                ]
                + [
                    {
                        "arm_name": t.arms[0].name,
                        "metric_name": "branin_b",
                        "mean": 0.4 * t.index,
                        "sem": 0.2 + 0.1 * t.index,
                        "trial_index": t.index,
                    }
                    for t in exp.trials.values()
                ]
            )
        )
        exp.attach_data(data)
        # Extract experiment data.
        experiment_data = extract_experiment_data(
            experiment=exp, data_loader_config=DataLoaderConfig()
        )
        # Arm data has been tested above, just checking observation data here.
        expected_obs_data = DataFrame(
            data=[
                [0.0, 0.0, 0.2, 0.2, Timestamp(0.0), Timestamp(5.0)],
                [0.4, 0.4, 0.3, 0.3, Timestamp(1.0), Timestamp(6.0)],
                [0.8, 0.8, 0.4, 0.4, Timestamp(2.0), Timestamp(7.0)],
            ],
            index=MultiIndex.from_tuples(
                [(0, "0_0"), (1, "1_0"), (2, "2_0")],
                names=["trial_index", "arm_name"],
            ),
            columns=MultiIndex.from_tuples(
                tuples=[
                    ("mean", "branin_a"),
                    ("mean", "branin_b"),
                    ("sem", "branin_a"),
                    ("sem", "branin_b"),
                    ("metadata", "start_time"),
                    ("metadata", "end_time"),
                ]
            ),
        )
        assert_frame_equal(expected_obs_data, experiment_data.observation_data)

    def test_filter_by_arm_name(self) -> None:
        # This is a 2 objective experiment with 5 trials, 1 arm each.
        observations = [[0.1, 1.0], [0.2, 2.0], [0.3, 3.0], [0.4, 4.0], [0.5, 5.0]]
        exp = get_experiment_with_observations(observations=observations)
        experiment_data = extract_experiment_data(
            experiment=exp, data_loader_config=DataLoaderConfig()
        )
        self.assertEqual(len(experiment_data.arm_data), 5)
        self.assertEqual(len(experiment_data.observation_data), 5)
        # Filter to only include arms 1 & 3.
        arm_names = ["1_0", "3_0"]
        filtered = experiment_data.filter_by_arm_names(arm_names=arm_names)
        self.assertEqual(
            list(filtered.arm_data.index.get_level_values("arm_name")), arm_names
        )
        self.assertEqual(
            list(filtered.observation_data.index.get_level_values("arm_name")),
            arm_names,
        )
        # Check that filtering was applied correctly.
        mask = [False, True, False, True, False]
        assert_frame_equal(filtered.arm_data, experiment_data.arm_data.loc[mask])
        assert_frame_equal(
            filtered.observation_data, experiment_data.observation_data.loc[mask]
        )

    def test_filter_by_trial_index(self) -> None:
        # This is a 2 objective experiment with 5 trials, 1 arm each.
        observations = [[0.1, 1.0], [0.2, 2.0], [0.3, 3.0], [0.4, 4.0], [0.5, 5.0]]
        exp = get_experiment_with_observations(observations=observations)
        experiment_data = extract_experiment_data(
            experiment=exp, data_loader_config=DataLoaderConfig()
        )
        self.assertEqual(len(experiment_data.arm_data), 5)
        self.assertEqual(len(experiment_data.observation_data), 5)
        # Filter to only include trials 1 & 3.
        trial_indices = [1, 3]
        filtered = experiment_data.filter_by_trial_index(trial_indices=trial_indices)
        self.assertEqual(
            list(filtered.arm_data.index.get_level_values("trial_index")), trial_indices
        )
        self.assertEqual(
            list(filtered.observation_data.index.get_level_values("trial_index")),
            trial_indices,
        )
        # Check that filtering was applied correctly.
        mask = [False, True, False, True, False]
        assert_frame_equal(filtered.arm_data, experiment_data.arm_data.loc[mask])
        assert_frame_equal(
            filtered.observation_data, experiment_data.observation_data.loc[mask]
        )

    def test_filter_latest_observations(self) -> None:
        exp = get_branin_experiment_with_timestamp_map_metric(with_trials_and_data=True)
        experiment_data = extract_experiment_data(
            experiment=exp,
            data_loader_config=DataLoaderConfig(
                fit_only_completed_map_metrics=False,
                latest_rows_per_group=None,
            ),
        )
        filtered_data = experiment_data.filter_latest_observations()
        self.assertNotEqual(filtered_data, experiment_data)
        # Arm data is the same.
        assert_frame_equal(experiment_data.arm_data, filtered_data.arm_data)
        # Observation data is filtered to only include one row for each arm
        # and no timestamp on the index.
        # In this case, the data is identical to timestamp 0 for branin_map and
        # timestamp nan for branin.

        # the first two rows of experiment_data.observation_data.loc[(0, "0_0")]:
        #                 mean               sem            metadata
        #               branin branin_map branin branin_map        n frac_nonnull
        # timestamp
        # NaN        55.602113        NaN    0.0        NaN  10000.0    55.602113
        # 0.0              NaN  55.602113    NaN        0.0      NaN          NaN

        # this operation coalesces it to take the non-null values, putting data
        # for branin and branin_map on the same row.
        expected_obs_data = (
            experiment_data.observation_data.loc[
                experiment_data.observation_data.index.get_level_values("step").isin(
                    (0, nan)
                )
            ]
            .reset_index("step", drop=True)
            .groupby(level=["trial_index", "arm_name"])
            .first()
        )

        assert_frame_equal(filtered_data.observation_data, expected_obs_data)

    def test_convert_to_list_of_observations(self) -> None:
        exp = get_branin_experiment_with_timestamp_map_metric(with_trials_and_data=True)
        # Complete the first trial, so that it has metadata.
        exp.trials[0].mark_completed()
        experiment_data = extract_experiment_data(
            experiment=exp,
            data_loader_config=DataLoaderConfig(
                fit_only_completed_map_metrics=False,
                latest_rows_per_group=None,
            ),
        )
        # With map keys.
        copy_data = deepcopy(experiment_data)
        observations = experiment_data.convert_to_list_of_observations()
        # Experiment data shouldn't be modified.
        self.assertEqual(experiment_data, copy_data)
        # Check that the observations are correct.
        branin_trial_0 = Observation(
            features=ObservationFeatures(
                parameters={"x1": 0.0, "x2": 0.0},
                trial_index=0,
                metadata={
                    Keys.TRIAL_COMPLETION_TIMESTAMP: mock.ANY,
                    "step": float("NaN"),
                },
            ),
            data=ObservationData(
                metric_names=["branin"],
                means=np.array([55.602112642270264]),
                covariance=np.diag([0.0]),
            ),
            arm_name="0_0",
        )
        branin_map_trial_0 = [
            Observation(
                features=ObservationFeatures(
                    parameters={"x1": 0.0, "x2": 0.0},
                    trial_index=0,
                    metadata={
                        Keys.TRIAL_COMPLETION_TIMESTAMP: mock.ANY,
                        "step": timestamp,
                    },
                ),
                data=ObservationData(
                    metric_names=["branin_map"],
                    means=np.array([55.602112642270264]),
                    covariance=np.diag([0.0]),
                ),
                arm_name="0_0",
            )
            for timestamp in [0.0, 1.0, 2.0, 3.0]
        ]
        branin_trial_1 = Observation(
            features=ObservationFeatures(
                parameters={"x1": 1.0, "x2": 1.0},
                trial_index=1,
                metadata={"step": float("NaN")},
            ),
            data=ObservationData(
                metric_names=["branin"],
                means=np.array([27.702905548512433]),
                covariance=np.diag([0.0]),
            ),
            arm_name="1_0",
        )
        branin_map_trial_1 = [
            Observation(
                features=ObservationFeatures(
                    parameters={"x1": 1.0, "x2": 1.0},
                    trial_index=1,
                    metadata={"step": timestamp},
                ),
                data=ObservationData(
                    metric_names=["branin_map"],
                    means=np.array([27.702905548512433]),
                    covariance=np.diag([0.0]),
                ),
                arm_name="1_0",
            )
            for timestamp in [0.0, 1.0]
        ]

        expected = (
            [branin_trial_0]
            + branin_map_trial_0
            + [branin_trial_1]
            + branin_map_trial_1
        )
        self.assertEqual(observations, expected)

        # After removing the map keys.
        experiment_data = experiment_data.filter_latest_observations()
        copy_data = deepcopy(experiment_data)
        observations = experiment_data.convert_to_list_of_observations()
        # Experiment data shouldn't be modified.
        self.assertEqual(experiment_data, copy_data)
        # Check that the observations are correct.
        expected = [
            Observation(
                features=ObservationFeatures(
                    parameters={"x1": 0.0, "x2": 0.0},
                    trial_index=0,
                    metadata={Keys.TRIAL_COMPLETION_TIMESTAMP: mock.ANY},
                ),
                data=ObservationData(
                    metric_names=["branin", "branin_map"],
                    # Means are deterministic based on the parameterization.
                    means=np.array([55.602112642270264, 55.602112642270264]),
                    covariance=np.diag([0.0, 0.0]),
                ),
                arm_name="0_0",
            ),
            Observation(
                features=ObservationFeatures(
                    parameters={"x1": 1.0, "x2": 1.0},
                    trial_index=1,
                    metadata={},
                ),
                data=ObservationData(
                    metric_names=["branin", "branin_map"],
                    means=np.array([27.702905548512433, 27.702905548512433]),
                    covariance=np.diag([0.0, 0.0]),
                ),
                arm_name="1_0",
            ),
        ]
        self.assertEqual(observations, expected)

    def test_experiment_data_metric_names(self) -> None:
        for experiment, expected in [
            (get_branin_experiment(), []),
            (get_branin_experiment(with_completed_trial=True), ["branin"]),
            (
                get_branin_experiment(
                    with_completed_trial=True, with_absolute_constraint=True
                ),
                ["branin", "branin_e"],
            ),
        ]:
            experiment_data = extract_experiment_data(
                experiment=experiment, data_loader_config=DataLoaderConfig()
            )
            self.assertEqual(experiment_data.metric_names, expected)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import re

DOT_PLACEHOLDER = "__dot__"
SLASH_PLACEHOLDER = "__slash__"
COLON_PLACEHOLDER = "__colon__"
_forbidden_re: re.Pattern[str] = re.compile(r"[\;\[\'\\]")


def sanitize_name(s: str) -> str:
    """
    Converts a string with normal dots and slashes to a string with sanitized dots and
    slashes. This is temporarily necessary because SymPy symbol names must be valid
    Python identifiers, but some legacy Ax users may include dots or slashes in their
    parameter names.

    Note that we need to be careful here not to sanitize the string too much, as some
    dots are meaningful (ex. the objective "foo.bar + 0.1 * baz" should be parsed as
    "foo__dot__bar + 0.1 * baz" not "foo__dot__bar + 0__dot__1 * baz").


    This does not allow obvious attack vectors  `;`, `[`, backslashes, and quotations.
    Colons, dots, and slashes are sanitized.
    """
    if _forbidden_re.search(s) is not None:
        raise ValueError(f"Expression {s} has forbidden control characters.")
    # Replace occurances of "." and "/" when they appear after a valid Python variable
    # name and before any alphanumeric character.
    sans_dots = re.sub(
        r"([a-zA-Z_][a-zA-Z0-9_]*)\.([a-zA-Z0-9_])",
        rf"\1{DOT_PLACEHOLDER}\2",
        s,
    )
    sans_slash = re.sub(
        r"([a-zA-Z_][a-zA-Z0-9_]*)\/([a-zA-Z0-9_])",
        rf"\1{SLASH_PLACEHOLDER}\2",
        sans_dots,
    )
    sans_colon = re.sub(
        r"([a-zA-Z_][a-zA-Z0-9_]*):([a-zA-Z0-9_])",
        rf"\1{COLON_PLACEHOLDER}\2",
        sans_slash,
    )

    return sans_colon


def unsanitize_name(s: str) -> str:
    """
    Converts a string with sanitized dots and slashes back to a string with normal dots
    and slashes. This is temporarily necessary because SymPy symbol names must be valid
    Python identifiers, but some legacy Ax users may include dots or slashes in their
    parameter names.
    """

    # Unsanitize in the reverse order of sanitization
    with_colon = re.sub(rf"{COLON_PLACEHOLDER}", ":", s)
    with_slash = re.sub(rf"{SLASH_PLACEHOLDER}", "/", with_colon)
    with_dot = re.sub(rf"{DOT_PLACEHOLDER}", ".", with_slash)

    return with_dot



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from collections.abc import Iterable

from ax.core.metric import Metric
from ax.exceptions.core import UserInputError
from ax.utils.common.base import SortableBase
from pyre_extensions import none_throws


class Objective(SortableBase):
    """Base class for representing an objective.

    Attributes:
        minimize: If True, minimize metric.
    """

    def __init__(self, metric: Metric, minimize: bool | None = None) -> None:
        """Create a new objective.

        Args:
            metric: The metric to be optimized.
            minimize: If True, minimize metric. If None, will be set based on the
                `lower_is_better` property of the metric (if that is not specified,
                will raise a `UserInputError`).

        """
        lower_is_better = metric.lower_is_better
        if minimize is None:
            if lower_is_better is None:
                raise UserInputError(
                    f"Metric {metric.name} does not specify `lower_is_better` "
                    "and `minimize` is not specified. At least one of these "
                    "must be specified."
                )
            else:
                minimize = lower_is_better
        elif lower_is_better is not None and lower_is_better != minimize:
            raise UserInputError(
                f"Metric {metric.name} specifies {lower_is_better=}, "
                "which doesn't match the specified optimization direction "
                f"{minimize=}."
            )
        self._metric: Metric = metric
        self.minimize: bool = none_throws(minimize)

    @property
    def metric(self) -> Metric:
        """Get the objective metric."""
        return self._metric

    @property
    def metrics(self) -> list[Metric]:
        """Get a list of objective metrics."""
        return [self._metric]

    @property
    def metric_names(self) -> list[str]:
        """Get a list of objective metric names."""
        return [m.name for m in self.metrics]

    def clone(self) -> Objective:
        """Create a copy of the objective."""
        return Objective(self.metric.clone(), self.minimize)

    def __repr__(self) -> str:
        return 'Objective(metric_name="{}", minimize={})'.format(
            self.metric.name, self.minimize
        )

    def get_unconstrainable_metrics(self) -> list[Metric]:
        """Return a list of metrics that are incompatible with OutcomeConstraints."""
        return self.metrics

    @property
    def _unique_id(self) -> str:
        return str(self)


class MultiObjective(Objective):
    """Class for an objective composed of a multiple component objectives.

    The Acquisition function determines how the objectives are weighted.

    Attributes:
        objectives: List of objectives.
    """

    def __init__(self, objectives: list[Objective]) -> None:
        """Create a new objective.

        Args:
            objectives: The list of objectives to be jointly optimized.

        """
        self._objectives: list[Objective] = objectives

    @property
    def metric(self) -> Metric:
        """Override base method to error."""
        raise NotImplementedError(
            f"{type(self).__name__} is composed of multiple metrics"
        )

    @property
    def metrics(self) -> list[Metric]:
        """Get the objective metrics."""
        return [o.metric for o in self._objectives]

    @property
    def objectives(self) -> list[Objective]:
        """Get the objectives."""
        return self._objectives

    def clone(self) -> MultiObjective:
        """Create a copy of the objective."""
        return MultiObjective(objectives=[o.clone() for o in self.objectives])

    def __repr__(self) -> str:
        return f"MultiObjective(objectives={self.objectives})"


class ScalarizedObjective(Objective):
    """Class for an objective composed of a linear scalarization of metrics.

    Attributes:
        metrics: List of metrics.
        weights: Weights for scalarization; default to 1.
    """

    weights: list[float]

    def __init__(
        self,
        metrics: list[Metric],
        weights: list[float] | None = None,
        minimize: bool = False,
    ) -> None:
        """Create a new objective.

        Args:
            metric: The metric to be optimized.
            weights: The weights for the linear combination of metrics.
            minimize: If true, minimize the linear combination.

        """
        if weights is None:
            weights = [1.0 for _ in metrics]
        else:
            if len(weights) != len(metrics):
                raise ValueError("Length of weights must equal length of metrics")

        # Check if the optimization direction is consistent with
        # `lower_is_better` (if specified).
        for m, w in zip(metrics, weights):
            is_minimized = minimize if w > 0 else not minimize
            if m.lower_is_better is not None and is_minimized != m.lower_is_better:
                raise ValueError(
                    f"Metric with name {m.name} specifies `lower_is_better` = "
                    f"{m.lower_is_better}, which doesn't match the specified "
                    "optimization direction. You most likely want to flip the sign of "
                    "the corresponding metric weight."
                )

        self._metrics = metrics
        self.weights = weights
        self.minimize = minimize

    @property
    def metric(self) -> Metric:
        """Override base method to error."""
        raise NotImplementedError(
            f"{type(self).__name__} is composed of multiple metrics"
        )

    @property
    def metrics(self) -> list[Metric]:
        """Get the metrics."""
        return self._metrics

    @property
    def metric_weights(self) -> Iterable[tuple[Metric, float]]:
        """Get the metrics and weights."""
        return zip(self.metrics, self.weights)

    def clone(self) -> ScalarizedObjective:
        """Create a copy of the objective."""
        return ScalarizedObjective(
            metrics=[m.clone() for m in self.metrics],
            weights=self.weights.copy(),
            minimize=self.minimize,
        )

    def __repr__(self) -> str:
        return "ScalarizedObjective(metric_names={}, weights={}, minimize={})".format(
            [metric.name for metric in self.metrics], self.weights, self.minimize
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Iterable

from ax.analysis.analysis import Analysis, display_cards
from ax.analysis.analysis_card import AnalysisCardBase
from ax.analysis.overview import OverviewAnalysis
from ax.core.experiment import Experiment
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.service.utils.with_db_settings_base import WithDBSettingsBase


class AnalysisBase(WithDBSettingsBase):
    """
    Base class for analysis functionality shared between AxClient and orchestrator.
    """

    # pyre-fixme[13]: Attribute `experiment` is declared in class
    # `AnalysisBase` to have type `Experiment` but is never initialized
    experiment: Experiment
    # pyre-fixme[13]: Attribute `generation_strategy` is declared in class
    # `AnalysisBase` to have type `GenerationStrategy` but
    # is never initialized
    generation_strategy: GenerationStrategy

    def compute_analyses(
        self,
        analyses: Iterable[Analysis] | None = None,
        display: bool = True,
    ) -> list[AnalysisCardBase]:
        """
        Compute AnalysisCards (data about the optimization for end-user consumption)
        using the Experiment and GenerationStrategy. If no analyses are provided use
        some heuristic to determine which analyses to run. If some analyses fail, log
        failure and continue to compute the rest.

        Note that the Analysis class is NOT part of the API and its methods are subject
        to change incompatibly between minor versions. Users are encouraged to use the
        provided analyses or leave this argument as None to use the default analyses.

        Saves to database on completion if storage_config is present.

        Args:
            analyses: A list of Analysis classes to run. If None Ax will choose which
                analyses to run based on the state of the experiment.
            display: Whether to display the AnalysisCards if executed in an interactive
                environment (e.g. Jupyter). Defaults to True. If not in an interactive
                environment this setting has no effect.
        Returns:
            A list of AnalysisCards.
        """

        analyses = analyses if analyses is not None else [OverviewAnalysis()]

        # Compute Analyses one by one and accumulate Results holding either the
        # AnalysisCard or an Exception and some metadata
        cards = [
            analysis.compute_or_error_card(
                experiment=self.experiment,
                generation_strategy=self.generation_strategy,
            )
            for analysis in analyses
        ]

        # Display the AnalysisCards if requested and if the user is in a notebook
        if display:
            display_cards(cards=cards)

        return cards
